from __future__ import annotations

import contextlib
import dataclasses
import importlib
import inspect
import os
from abc import ABCMeta
from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime, tzinfo
from functools import cache
from types import FunctionType, WrapperDescriptorType
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Protocol,
    TypeVar,
    cast,
    get_args,
    get_origin,
    overload,
    runtime_checkable,
)
from uuid import UUID, uuid4
from warnings import warn

from eventsourcing.utils import (
    TopicError,
    get_method_name,
    get_topic,
    register_topic,
    resolve_topic,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from typing_extensions import Self


TZINFO: tzinfo = resolve_topic(os.getenv("TZINFO_TOPIC", "datetime:timezone.utc"))
"""
A Python :py:obj:`tzinfo` object that defaults to UTC (:py:obj:`timezone.utc`). Used
as the timezone argument in :func:`~eventsourcing.domain.datetime_now_with_tzinfo`.

Set environment variable ``TZINFO_TOPIC`` to the topic of a different :py:obj:`tzinfo`
object so that all your domain model event timestamps are located in that timezone
(not recommended). It is generally recommended to locate all timestamps in the UTC
domain and convert to local timezones when presenting values in user interfaces.
"""


class EventsourcingType(type):
    """Base type for event sourcing domain model types (aggregates and events)."""


_T = TypeVar("_T")


def patch_dataclasses_process_class() -> None:
    dataclasses_module = importlib.import_module("dataclasses")
    original_process_class_func = dataclasses_module.__dict__["_process_class"]

    def _patched_dataclasses_process_class(
        cls: type[_T], *args: Any, **kwargs: Any
    ) -> type[_T]:
        # Avoid processing aggregate and event dataclasses twice,
        # because doing so screws up non-init and default fields.
        if (
            cls
            and isinstance(cls, EventsourcingType)
            and "__dataclass_fields__" in cls.__dict__
        ):
            return cls
        return original_process_class_func(cls, *args, **kwargs)

    dataclasses_module.__dict__["_process_class"] = _patched_dataclasses_process_class


patch_dataclasses_process_class()


TAggregateID = TypeVar("TAggregateID", bound=UUID | str)
TAggregateID_co = TypeVar("TAggregateID_co", bound=UUID | str, covariant=True)


@runtime_checkable
class DomainEventProtocol(Protocol[TAggregateID_co]):
    """Protocol for domain event objects.

    A protocol is defined to allow the event sourcing mechanisms
    to work with different kinds of domain event classes. Whilst
    the library by default uses frozen dataclasses to implement
    its domain event classes, it is also possible to use other
    kinds of domain event classes, such as Pydantic classes.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass  # pragma: no cover

    @property
    def originator_id(self) -> TAggregateID_co:
        """UUID identifying an aggregate to which the event belongs."""
        raise NotImplementedError  # pragma: no cover

    @property
    def originator_version(self) -> int:
        """Integer identifying the version of the aggregate when the event occurred."""
        raise NotImplementedError  # pragma: no cover


TDomainEvent = TypeVar("TDomainEvent", bound=DomainEventProtocol[Any])
SDomainEvent = TypeVar("SDomainEvent", bound=DomainEventProtocol[Any])


class MutableAggregateProtocol(Protocol[TAggregateID_co]):
    """Protocol for mutable aggregate objects.

    A protocol is defined to allow the event sourcing mechanisms
    to work with different kinds of aggregate classes. Whilst
    the library by default recommends using mutable classes to
    implement aggregate classes, it is also possible to implement
    immutable aggregate classes, and this is supported by this library.
    """

    @property
    def id(self) -> TAggregateID_co:
        """Mutable aggregates have a read-only ID that is a UUID."""
        raise NotImplementedError  # pragma: no cover

    @property
    def version(self) -> int:
        """Mutable aggregates have a read-write version that is an int."""
        raise NotImplementedError  # pragma: no cover

    @version.setter
    def version(self, value: int) -> None:
        """Mutable aggregates have a read-write version that is an int."""
        raise NotImplementedError  # pragma: no cover


class ImmutableAggregateProtocol(Protocol[TAggregateID_co]):
    """Protocol for immutable aggregate objects.

    A protocol is defined to allow the event sourcing mechanisms
    to work with different kinds of aggregate classes. Whilst
    the library by default recommends using mutable classes to
    implement aggregate classes, it is also possible to implement
    immutable aggregate classes, and this is supported by this library.
    """

    @property
    def id(self) -> TAggregateID_co:
        """Immutable aggregates have a read-only ID that is a UUID."""
        raise NotImplementedError  # pragma: no cover

    @property
    def version(self) -> int:
        """Immutable aggregates have a read-only version that is an int."""
        raise NotImplementedError  # pragma: no cover


MutableOrImmutableAggregate = (
    ImmutableAggregateProtocol[TAggregateID] | MutableAggregateProtocol[TAggregateID]
)
"""Type alias defining a union of mutable and immutable aggregate protocols."""


TMutableOrImmutableAggregate = TypeVar(
    "TMutableOrImmutableAggregate", bound=MutableOrImmutableAggregate[Any]
)
"""Type variable bound by the union of mutable and immutable aggregate protocols."""


@runtime_checkable
class CollectEventsProtocol(Protocol):
    """Protocol for aggregates that support collecting pending events."""

    def collect_events(self) -> Sequence[DomainEventProtocol[Any]]:
        """Returns a sequence of events."""
        raise NotImplementedError  # pragma: no cover


@runtime_checkable
class CanMutateProtocol(
    DomainEventProtocol[Any], Protocol[TMutableOrImmutableAggregate]
):
    """Protocol for events that have a mutate method."""

    def mutate(
        self, aggregate: TMutableOrImmutableAggregate | None
    ) -> TMutableOrImmutableAggregate | None:
        """
        Evolves the state of the given aggregate, either by
        returning the given aggregate instance with modified attributes
        or by constructing and returning a new aggregate instance.
        """


def datetime_now_with_tzinfo() -> datetime:
    """
    Constructs a timezone-aware :class:`datetime`
    object for the current date and time.

    Uses :py:obj:`TZINFO` as the timezone.
    """
    return datetime.now(tz=TZINFO)


def create_utc_datetime_now() -> datetime:
    """Deprected in favour of :func:`~eventsourcing.domain.datetime_now_with_tzinfo`."""
    msg = (
        "'create_utc_datetime_now()' is deprecated, "
        "use 'datetime_now_with_tzinfo()' instead"
    )
    warn(msg, DeprecationWarning, stacklevel=2)
    return datetime_now_with_tzinfo()


class CanCreateTimestamp:
    """Provides a create_timestamp() method to subclasses."""

    @staticmethod
    def create_timestamp() -> datetime:
        """Constructs a timezone-aware :class:`datetime` object
        representing when an event occurred.
        """
        return datetime_now_with_tzinfo()


TAggregate = TypeVar("TAggregate", bound="BaseAggregate[Any]")


class AbstractDecision:
    pass


class HasOriginatorIDVersion(AbstractDecision, Generic[TAggregateID]):
    """Declares ``originator_id`` and ``originator_version`` attributes."""

    originator_id: TAggregateID
    """UUID identifying an aggregate to which the event belongs."""
    originator_version: int
    """Integer identifying the version of the aggregate when the event occurred."""

    originator_id_type: ClassVar[type[UUID | str] | None] = None

    def __init_subclass__(cls) -> None:
        cls.find_originator_id_type(HasOriginatorIDVersion)
        super().__init_subclass__()

    @classmethod
    def find_originator_id_type(cls: type, generic_cls: type) -> None:
        """Store the type argument of TAggregateID on the subclass."""
        if "originator_id_type" not in cls.__dict__:
            for orig_base in cls.__orig_bases__:  # type: ignore[attr-defined]
                if "originator_id_type" in orig_base.__dict__:
                    cls.originator_id_type = orig_base.__dict__["originator_id_type"]  # type: ignore[attr-defined]
                elif get_origin(orig_base) is generic_cls:
                    originator_id_type = get_args(orig_base)[0]
                    if originator_id_type in (UUID, str):
                        cls.originator_id_type = originator_id_type  # type: ignore[attr-defined]
                        break
                    if originator_id_type is Any:
                        continue
                    if isinstance(originator_id_type, TypeVar):
                        continue
                    msg = f"Aggregate ID type arg cannot be {originator_id_type}"
                    raise TypeError(msg)


class CanMutateAggregate(HasOriginatorIDVersion[TAggregateID], CanCreateTimestamp):
    """Implements a :py:func:`~eventsourcing.domain.CanMutateAggregate.mutate`
    method that evolves the state of an aggregate.
    """

    # TODO: Move this to a HasTimestamp? Why is it here??
    timestamp: datetime
    """Timezone-aware :class:`datetime` object representing when an event occurred."""

    def __init_subclass__(cls) -> None:
        cls.find_originator_id_type(CanMutateAggregate)
        super().__init_subclass__()

    def mutate(self, aggregate: TAggregate | None) -> TAggregate | None:
        """Validates and adjusts the attributes of the given ``aggregate``
         argument. The argument is typed as ``Optional``, but the value is
         expected to be not ``None``.

        Validates the ``aggregate`` argument by checking the event's
        :py:attr:`~eventsourcing.domain.HasOriginatorIDVersion.originator_id` equals the
        ``aggregate``'s :py:attr:`~eventsourcing.domain.Aggregate.id`, and the event's
        :py:attr:`~eventsourcing.domain.HasOriginatorIDVersion.originator_version` is
        one greater than the ``aggregate``'s current
        :py:attr:`~eventsourcing.domain.Aggregate.version`.
        If the ``aggregate`` argument is not valid, an exception is raised.

        If the ``aggregate`` argument is valid, the
        :func:`~eventsourcing.domain.CanMutateAggregate.apply` method is called, and
        then :py:attr:`~eventsourcing.domain.HasOriginatorIDVersion.originator_id` is
        assigned to the aggregate's :py:attr:`~eventsourcing.domain.Aggregate.version`
        attribute, and the ``timestamp`` is assigned to the aggregate's
        :py:attr:`~eventsourcing.domain.Aggregate.modified_on` attribute.
        """
        assert aggregate is not None

        # Check this event belongs to this aggregate.
        if self.originator_id != aggregate.id:
            raise OriginatorIDError(self.originator_id, aggregate.id)

        # Check this event is the next in its sequence.
        next_version = aggregate.version + 1
        if self.originator_version != next_version:
            raise OriginatorVersionError(self.originator_version, next_version)

        # Call apply() before mutating values, in case exception is raised.
        self.apply(aggregate)

        # Update the aggregate's 'version' number.
        aggregate.version = self.originator_version

        # Update the aggregate's 'modified on' time.
        aggregate.modified_on = self.timestamp

        # Return the mutated aggregate.
        return aggregate

    def apply(self, aggregate: Any) -> None:
        """Applies the domain event to its aggregate.

        This method does nothing but exist to be
        overridden as a convenient way for users
        to define how an event evolves the state
        of an aggregate.
        """

    def _as_dict(self) -> dict[str, Any]:
        return vars(self)


class CanInitAggregate(CanMutateAggregate[TAggregateID]):
    """Implements a :func:`~eventsourcing.domain.CanMutateAggregate.mutate`
    method that constructs the initial state of an aggregate.
    """

    originator_topic: str
    """String describing the path to an aggregate class."""

    def __init_subclass__(cls) -> None:
        cls.find_originator_id_type(CanInitAggregate)
        super().__init_subclass__()

    def mutate(self, aggregate: TAggregate | None) -> TAggregate | None:
        """Constructs an aggregate instance according to the attributes of an event.

        The ``aggregate`` argument is typed as an optional argument, but the
        value is expected to be ``None``.
        """
        assert aggregate is None

        # Resolve originator topic.
        aggregate_class: type[TAggregate] = resolve_topic(self.originator_topic)

        # Construct an aggregate object (a "shell" of the correct object type).
        agg = aggregate_class.__new__(aggregate_class)

        # Pick out event attributes for the aggregate base class init method.
        self_dict = self._as_dict()
        base_kwargs = filter_kwargs_for_method_params(
            self_dict, type(agg).__base_init__
        )

        # Call the base class init method (so we don't need to always write
        # a call to super().__init__() in every aggregate __init__() method).
        agg.__base_init__(**base_kwargs)

        # Pick out event attributes for aggregate subclass class init method.
        init_kwargs = filter_kwargs_for_method_params(self_dict, type(agg).__init__)

        # Provide the aggregate id, if the __init__ method expects it.
        if aggregate_class in _init_mentions_id:
            init_kwargs["id"] = self_dict["originator_id"]

        # Call the aggregate subclass class init method.
        agg.__init__(**init_kwargs)  # type: ignore[misc]

        # Call the event apply method (alternative to using __init__())
        self.apply(agg)

        # Return the constructed and initialised aggregate object.
        return agg


class MetaDomainEvent(EventsourcingType):
    """Metaclass which ensures all domain event classes are frozen dataclasses."""

    def __new__(
        cls, name: str, bases: tuple[type[TDomainEvent], ...], cls_dict: dict[str, Any]
    ) -> type[TDomainEvent]:
        event_cls = cast(
            "type[TDomainEvent]", super().__new__(cls, name, bases, cls_dict)
        )
        event_cls = dataclasses.dataclass(frozen=True)(event_cls)
        event_cls.__signature__ = inspect.signature(event_cls.__init__)  # type: ignore[attr-defined]
        return event_cls


@dataclass(frozen=True)
class DomainEvent(CanCreateTimestamp, metaclass=MetaDomainEvent):
    """Frozen data class representing domain model events."""

    originator_id: UUID
    """UUID identifying an aggregate to which the event belongs."""
    originator_version: int
    """Integer identifying the version of the aggregate when the event occurred."""
    timestamp: datetime
    """Timezone-aware :class:`datetime` object representing when an event occurred."""

    def __post_init__(self) -> None:
        if not isinstance(self.originator_id, UUID):
            msg = (
                f"{type(self)} "
                f"was initialized with a non-UUID originator_id: "
                f"{self.originator_id!r}"
            )
            raise TypeError(msg)


class AggregateEvent(CanMutateAggregate[UUID], DomainEvent):
    """Frozen data class representing aggregate events.

    Subclasses represent original decisions made by domain model aggregates.
    """


@dataclass(frozen=True)
class AggregateCreated(CanInitAggregate[UUID], AggregateEvent):
    """Frozen data class representing the initial creation of an aggregate."""

    originator_topic: str
    """String describing the path to an aggregate class."""


class EventSourcingError(Exception):
    """Base exception class."""


class ProgrammingError(EventSourcingError):
    """Exception class for domain model programming errors."""


class LogEvent(DomainEvent):
    """Deprecated: Inherit from DomainEvent instead.

    Base class for the events of event-sourced logs.
    """


def filter_kwargs_for_method_params(
    kwargs: dict[str, Any], method: Callable[..., Any]
) -> dict[str, Any]:
    names = _spec_filter_kwargs_for_method_params(method)
    return {k: v for k, v in kwargs.items() if k in names}


@cache
def _spec_filter_kwargs_for_method_params(method: Callable[..., Any]) -> set[str]:
    method_signature = inspect.signature(method)
    return set(method_signature.parameters)


if TYPE_CHECKING:
    EventSpecType = str | type[AbstractDecision]

CallableType = Callable[..., None]
DecoratableType = CallableType | property
TDecoratableType = TypeVar("TDecoratableType", bound=DecoratableType)


class CommandMethodDecorator:
    def __init__(
        self,
        event_spec: EventSpecType | None,
        decorated_obj: DecoratableType,
        event_topic: str | None = None,
    ):
        self.is_name_inferred_from_method = False
        self.given_event_cls: (
            type[CanMutateAggregate[Any] | AbstractDecision] | None
        ) = None
        self.event_cls_name: str | None = None
        self.decorated_property: property | None = None
        self.is_property_setter = False
        self.property_setter_arg_name: str | None = None
        self.decorated_func: CallableType
        self.event_topic = event_topic

        # Event name has been specified.
        if isinstance(event_spec, str):
            if event_spec == "":
                msg = "Can't use empty string as name of event class"
                raise ValueError(msg)
            self.event_cls_name = event_spec

        # Event class has been specified.
        elif isinstance(event_spec, type) and issubclass(
            event_spec, (CanMutateAggregate, AbstractDecision)
        ):
            # Guard against associating more than one method body with any given class.
            if (
                issubclass(event_spec, CanMutateAggregate)
                and event_spec in _given_event_classes
            ):
                name = event_spec.__name__
                msg = f"{name} event class used in more than one decorator"
                raise TypeError(msg)
            self.given_event_cls = event_spec
            _given_event_classes.add(event_spec)

        # Process a decorated property.
        if isinstance(decorated_obj, property):
            # Disallow putting event decorator on property getter.
            if decorated_obj.fset is None:
                assert decorated_obj.fget, "Property has no getter"
                method_name = decorated_obj.fget.__name__
                msg = f"@event can't decorate {method_name}() property getter"
                raise TypeError(msg)

            # Remember we are decorating a property.
            self.decorated_property = decorated_obj

            # TODO: Disallow unusual property setters in more detail.
            assert isinstance(decorated_obj.fset, FunctionType)

            # Disallow deriving event class names from property names.
            if not self.given_event_cls and not self.event_cls_name:
                method_name = decorated_obj.fset.__name__
                msg = (
                    f"@event decorator on @{method_name}.setter "
                    f"requires event name or class"
                )
                raise TypeError(msg)

            # Remember property "setter" as the decorated function.
            self.decorated_func = decorated_obj.fset

            # Remember the name of the second setter arg.
            setter_arg_names = list(inspect.signature(self.decorated_func).parameters)
            assert len(setter_arg_names) == 2
            self.property_setter_arg_name = setter_arg_names[1]

        # Process a decorated function.
        elif isinstance(decorated_obj, FunctionType):
            # Remember the decorated obj as the decorated method.
            self.decorated_func = decorated_obj

            all_func_decorators.append(self)
            # If necessary, derive an event class name from the method.
            if not self.given_event_cls and not self.event_cls_name:
                original_method_name = self.decorated_func.__name__
                if original_method_name != "__init__":
                    self.is_name_inferred_from_method = True
                    self.event_cls_name = "".join(
                        [s.capitalize() for s in original_method_name.split("_")]
                    )

        # Disallow decorating other types of object.
        else:
            msg = f"{decorated_obj} is not a function or property"
            raise TypeError(msg)

        # Disallow using methods with variable params to define event class.
        if self.event_cls_name:
            _raise_type_error_if_func_has_variable_params(self.decorated_func)

        # Disallow using methods with positional only params to define event class.
        if self.event_cls_name:
            _raise_type_error_if_func_has_positional_only_params(self.decorated_func)

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        # Initialised decorator was called directly, presumably by
        # a decorating property that has this decorator as its fset.
        # So trigger an event.
        assert self.is_property_setter
        assert self.property_setter_arg_name
        assert len(args) == 2
        assert len(kwargs) == 0
        assert isinstance(args[0], Aggregate)
        aggregate_instance = args[0]
        bound = BoundCommandMethodDecorator(self, aggregate_instance)
        property_setter_arg_value = args[1]
        kwargs = {self.property_setter_arg_name: property_setter_arg_value}
        bound.trigger(**kwargs)

    @overload
    def __get__(
        self, instance: None, owner: type[BaseAggregate[Any]]
    ) -> UnboundCommandMethodDecorator | property:
        """
        Descriptor protocol for getting decorated method or property on class object.
        """

    @overload
    def __get__(
        self, instance: BaseAggregate[Any], owner: type[BaseAggregate[Any]]
    ) -> BoundCommandMethodDecorator | Any:
        """
        Descriptor protocol for getting decorated method or property on instance object.
        """

    def __get__(
        self, instance: BaseAggregate[Any] | None, owner: type[BaseAggregate[Any]]
    ) -> BoundCommandMethodDecorator | UnboundCommandMethodDecorator | property | Any:
        """Descriptor protocol for getting decorated method or property."""
        if self.decorated_func.__name__ == "_":
            msg = "Underscore 'non-command' methods cannot be used to trigger events."
            raise ProgrammingError(msg)

        # If we are decorating a property, then delegate to the property's __get__.
        if self.decorated_property:
            return self.decorated_property.__get__(instance, owner)

        # If we are decorating an __init__ method, then delegate to the __init__ method.
        if self.decorated_func.__name__ == "__init__":
            return self.decorated_func.__get__(instance, owner)

        # Return a "bound" command method decorator if we have an instance.
        if instance:
            return BoundCommandMethodDecorator(self, instance)

        if "SPHINX_BUILD" in os.environ:  # pragma: no cover
            # Sphinx hack: use the original function when sphinx is running so that the
            # documentation ends up with the correct function signatures.
            # See 'SPHINX_BUILD' in conf.py.
            return self.decorated_func

        # Return an "unbound" command method decorator if we have no instance.
        return UnboundCommandMethodDecorator(self)

    def __set__(self, instance: BaseAggregate[Any], value: Any) -> None:
        """Descriptor protocol for assigning to decorated property."""
        # Set decorated property indirectly by triggering an event.
        assert self.property_setter_arg_name
        b = BoundCommandMethodDecorator(self, instance)
        kwargs = {self.property_setter_arg_name: value}
        b.trigger(**kwargs)


@overload
def event(arg: TDecoratableType, /) -> TDecoratableType:
    """Signature for calling ``@event`` decorator with decorated method."""


@overload
def event(
    arg: type[CanMutateAggregate[Any] | AbstractDecision], /
) -> Callable[[TDecoratableType], TDecoratableType]:
    """Signature for calling ``@event`` decorator with event class."""


@overload
def event(
    arg: str, /, *, topic: str | None = None
) -> Callable[[TDecoratableType], TDecoratableType]:
    """Signature for calling ``@event`` decorator with event name."""


@overload
def event(arg: None = None, /) -> Callable[[TDecoratableType], TDecoratableType]:
    """Signature for calling ``@event`` decorator without event specification."""


def event(
    arg: EventSpecType | TDecoratableType | None = None, /, *, topic: str | None = None
) -> TDecoratableType | Callable[[TDecoratableType], TDecoratableType]:
    """Event-triggering decorator for aggregate command methods and property setters.

    Can be used to decorate an aggregate method or property setter so that an
    event will be triggered when the method is called or the property is set.
    The body of the method will be used to apply the event to the aggregate,
    both when the event is triggered and when the aggregate is reconstructed
    from stored events.

    .. code-block:: python

        class MyAggregate(Aggregate):
            @event("NameChanged")
            def set_name(self, name: str):
                self.name = name

    ...is equivalent to...

    .. code-block:: python

        class MyAggregate(Aggregate):
            def set_name(self, name: str):
                self.trigger_event(self.NameChanged, name=name)

            class NameChanged(Aggregate.Event):
                name: str

                def apply(self, aggregate):
                    aggregate.name = self.name

    In the example above, the event "NameChanged" is defined automatically
    by inspecting the signature of the ``set_name()`` method. If it is
    preferred to declare the event class explicitly, for example to define
    upcasting of old events, the event class itself can be mentioned in the
    event decorator rather than just providing the name of the event as a
    string.

    .. code-block:: python

        class MyAggregate(Aggregate):

            class NameChanged(Aggregate.Event):
                name: str

            @event(NameChanged)
            def set_name(self, name: str):
                aggregate.name = self.name


    """
    if isinstance(arg, (FunctionType, property)):
        command_method_decorator = CommandMethodDecorator(
            event_spec=None,
            decorated_obj=arg,
        )
        return cast(
            "Callable[[TDecoratableType], TDecoratableType]", command_method_decorator
        )

    if (
        arg is None
        or isinstance(arg, str)
        or (
            isinstance(arg, type)
            and issubclass(arg, (CanMutateAggregate, AbstractDecision))
        )
    ):
        event_spec = arg

        def create_command_method_decorator(
            decorated_obj: TDecoratableType,
        ) -> TDecoratableType:
            command_method_decorator = CommandMethodDecorator(
                event_spec=event_spec,
                decorated_obj=decorated_obj,
                event_topic=topic,
            )
            return cast("TDecoratableType", command_method_decorator)

        return create_command_method_decorator

    msg = (
        f"{arg} is not a str, function, property, or subclass of "
        f"{CanMutateAggregate.__name__}"
    )
    raise TypeError(msg)


triggers = event


class UnboundCommandMethodDecorator:
    """Wraps a CommandMethodDecorator instance when accessed on an aggregate class."""

    def __init__(self, event_decorator: CommandMethodDecorator):
        """:param CommandMethodDecorator event_decorator:"""
        self.event_decorator = event_decorator
        self.__module__ = event_decorator.decorated_func.__module__
        self.__name__ = event_decorator.decorated_func.__name__
        self.__qualname__ = event_decorator.decorated_func.__qualname__
        self.__annotations__ = event_decorator.decorated_func.__annotations__
        self.__doc__ = event_decorator.decorated_func.__doc__
        # self.__wrapped__ = event_decorator.decorated_method
        # functools.update_wrapper(self, event_decorator.decorated_method)

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        # TODO: Review this, because other subclasses of BaseAggregate might too....
        # Expect first argument is an aggregate instance.
        if len(args) < 1 or not isinstance(args[0], Aggregate):
            msg = "Expected aggregate as first argument"
            raise TypeError(msg)
        aggregate: Aggregate = args[0]
        assert isinstance(aggregate, Aggregate)
        BoundCommandMethodDecorator(self.event_decorator, aggregate)(
            *args[1:], **kwargs
        )


class CanTriggerEvent(Protocol):
    def trigger_event(
        self,
        event_class: type[Any],
        **kwargs: Any,
    ) -> None:
        pass  # pragma: no cover


class BoundCommandMethodDecorator:
    """Binds a CommandMethodDecorator with an object instance that can trigger
    events, so that calls to decorated command methods can be intercepted and
    will trigger a "decorated func caller" event.
    """

    def __init__(self, event_decorator: CommandMethodDecorator, obj: CanTriggerEvent):
        """:param CommandMethodDecorator event_decorator:
        :param Aggregate aggregate:
        """
        self.event_decorator = event_decorator
        self.__module__ = event_decorator.decorated_func.__module__
        self.__name__ = event_decorator.decorated_func.__name__
        self.__qualname__ = event_decorator.decorated_func.__qualname__
        self.__annotations__ = event_decorator.decorated_func.__annotations__
        self.__doc__ = event_decorator.decorated_func.__doc__
        self.obj = obj

    def trigger(self, *args: Any, **kwargs: Any) -> None:
        kwargs = _coerce_args_to_kwargs(
            self.event_decorator.decorated_func, args, kwargs
        )
        try:
            event_cls = decorated_func_callers[self.event_decorator]
        except KeyError as e:  # pragma: no cover
            msg = (
                f"Event class not registered for event decorator on "
                f"{self.event_decorator.decorated_func.__qualname__}"
            )
            raise KeyError(msg) from e
        kwargs = filter_kwargs_for_method_params(kwargs, event_cls)
        self.obj.trigger_event(event_cls, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> None:
        self.trigger(*args, **kwargs)


class AbstractDecoratedFuncCaller:
    pass


class DecoratedFuncCaller(CanMutateAggregate[Any], AbstractDecoratedFuncCaller):
    def apply(self, aggregate: BaseAggregate[Any]) -> None:
        """Applies event to aggregate by calling method decorated by @event."""
        # Identify the function that was decorated.
        decorated_func = decorated_funcs[type(self)]

        # Select event attributes mentioned in function signature.
        self_dict = self._as_dict()
        kwargs = filter_kwargs_for_method_params(self_dict, decorated_func)

        # Call the original method with event attribute values.
        decorated_method = decorated_func.__get__(aggregate, type(aggregate))
        decorated_method(**kwargs)

        # Call super method, just in case any base classes need it.
        super().apply(aggregate)


# This helps enforce single usage of original event classes in decorators.
_given_event_classes = set[type]()

# This keeps track of the "created" event classes for an aggregate.
_created_event_classes: dict[type, list[type[CanInitAggregate[Any]]]] = {}

# This remembers which event class to trigger when a decorated method is called.
decorated_func_callers: dict[CommandMethodDecorator, type[AbstractDecision]] = {}

# This remembers which decorated func a decorated func caller should call.
decorated_funcs: dict[type, CallableType] = {}

# This keeps track of decorators on "non-command" projection-only methods.
all_func_decorators: list[CommandMethodDecorator] = []


def _raise_type_error_if_func_has_variable_params(method: CallableType) -> None:
    for param in inspect.signature(method).parameters.values():
        if param.kind is param.VAR_POSITIONAL:
            msg = f"*{param.name} not supported by decorator on {method.__name__}()"
            raise TypeError(msg)
            # TODO: Support VAR_POSITIONAL?
            # annotations["__star_args__"] = "typing.Any"

        if param.kind is param.VAR_KEYWORD:
            # TODO: Support VAR_KEYWORD?
            # annotations["__star_kwargs__"] = "typing.Any"
            msg = f"**{param.name} not supported by decorator on {method.__name__}()"
            raise TypeError(msg)


def _raise_type_error_if_func_has_positional_only_params(method: CallableType) -> None:
    # TODO: Support POSITIONAL_ONLY?
    positional_only_params = []
    for param in inspect.signature(method).parameters.values():
        if param.name == "self":
            continue
        if param.kind is param.POSITIONAL_ONLY:
            positional_only_params.append(param.name)

    if positional_only_params:
        msg = (
            f"positional only args arg not supported by @event decorator on "
            f"{method.__name__}(): {', '.join(positional_only_params)}"
        )
        raise TypeError(msg)


def _coerce_args_to_kwargs(
    method: CallableType,
    args: Iterable[Any],
    kwargs: dict[str, Any],
    *,
    expects_id: bool = False,
) -> dict[str, Any]:
    # __init__ methods are WrapperDescriptorType, other method are FunctionType.
    assert isinstance(method, (FunctionType, WrapperDescriptorType)), method

    args = tuple(args)
    enumerated_args_names, keyword_defaults_items = _spec_coerce_args_to_kwargs(
        method=method,
        len_args=len(args),
        kwargs_keys=tuple(kwargs.keys()),
        expects_id=expects_id,
    )

    copy_kwargs = dict(kwargs)
    copy_kwargs.update({name: args[i] for i, name in enumerated_args_names})
    copy_kwargs.update(keyword_defaults_items)
    return copy_kwargs


@cache
def _spec_coerce_args_to_kwargs(
    method: CallableType,
    len_args: int,
    kwargs_keys: tuple[str],
    *,
    expects_id: bool,
) -> tuple[tuple[tuple[int, str], ...], tuple[tuple[str, Any], ...]]:
    method_signature = inspect.signature(method)
    positional_names = []
    keyword_defaults = {}
    required_positional = []
    required_keyword_only = []
    if expects_id:
        positional_names.append("id")
        required_positional.append("id")
    for name, param in method_signature.parameters.items():
        if name == "self":
            continue
        # elif param.kind in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD):
        if param.kind is param.KEYWORD_ONLY:
            required_keyword_only.append(name)
        if param.kind is param.POSITIONAL_OR_KEYWORD:
            positional_names.append(name)
            if param.default == param.empty:
                required_positional.append(name)
        if param.default != param.empty:
            keyword_defaults[name] = param.default
    # if not required_keyword_only and not positional_names:
    #     if args or kwargs:
    #         raise TypeError(f"{method.__name__}() takes no args")
    method_name = get_method_name(method)
    for name in kwargs_keys:
        if name not in required_keyword_only and name not in positional_names:
            msg = f"{method_name}() got an unexpected keyword argument '{name}'"
            raise TypeError(msg)
    if len_args > len(positional_names):
        msg = (
            f"{method_name}() takes {len(positional_names) + 1} "
            f"positional argument{'' if len(positional_names) + 1 == 1 else 's'} "
            f"but {len_args + 1} were given"
        )
        raise TypeError(msg)
    required_positional_not_in_kwargs = [
        n for n in required_positional if n not in kwargs_keys
    ]
    num_missing = len(required_positional_not_in_kwargs) - len_args
    if num_missing > 0:
        missing_names = [
            f"'{name}'" for name in required_positional_not_in_kwargs[len_args:]
        ]
        msg = (
            f"{method_name}() missing {num_missing} required positional "
            f"argument{'' if num_missing == 1 else 's'}: "
        )
        _raise_missing_names_type_error(missing_names, msg)
    args_names = []
    for counter, name in enumerate(positional_names):
        if counter + 1 > len_args:
            break
        if name in kwargs_keys:
            msg = f"{method_name}() got multiple values for argument '{name}'"
            raise TypeError(msg)
        args_names.append(name)
    missing_keyword_only_arguments = [
        name for name in required_keyword_only if name not in kwargs_keys
    ]
    if missing_keyword_only_arguments:
        missing_names = [f"'{name}'" for name in missing_keyword_only_arguments]
        msg = (
            f"{method_name}() missing {len(missing_names)} "
            "required keyword-only argument"
            f"{'' if len(missing_names) == 1 else 's'}: "
        )
        _raise_missing_names_type_error(missing_names, msg)
    for key in tuple(keyword_defaults.keys()):
        if key in args_names or key in kwargs_keys:
            keyword_defaults.pop(key)
    enumerated_args_names = tuple(enumerate(args_names))
    keyword_defaults_items = tuple(keyword_defaults.items())
    return enumerated_args_names, keyword_defaults_items


def _raise_missing_names_type_error(missing_names: list[str], msg: str) -> None:
    msg += missing_names[0]
    if len(missing_names) == 2:
        msg += f" and {missing_names[1]}"
    elif len(missing_names) > 2:
        msg += ", " + ", ".join(missing_names[1:-1])
        msg += f", and {missing_names[-1]}"
    raise TypeError(msg)


_annotations_mention_id: set[type[BaseAggregate[Any]]] = set()
_init_mentions_id: set[type[BaseAggregate[Any]]] = set()
_create_id_param_names: dict[type[BaseAggregate[Any]], list[str]] = defaultdict(list)

ENVVAR_DISABLE_REDEFINITION_CHECK = "EVENTSOURCING_DISABLE_REDEFINITION_CHECK"


class MetaAggregate(EventsourcingType, ABCMeta, Generic[TAggregate]):
    """Metaclass for aggregate classes."""

    def _define_event_class(
        cls,
        name: str,
        bases: tuple[type[CanMutateAggregate[Any]], ...],
        apply_method: CallableType | None,
        event_topic: str | None = None,
    ) -> type[CanMutateAggregate[Any]]:
        # Define annotations for the event class (specs the init method).
        annotations = {}
        if apply_method is not None:
            method_signature = inspect.signature(apply_method)
            supers = {
                s for b in bases for s in b.__mro__ if hasattr(s, "__annotations__")
            }
            super_annotations = {a for s in supers for a in s.__annotations__}
            for param_name, param in list(method_signature.parameters.items())[1:]:
                # Don't define 'id' on a "created" class.
                if param_name == "id" and apply_method.__name__ == "__init__":
                    continue
                # Don't override super class annotations, unless no default on param.
                if param_name not in super_annotations or param.default == param.empty:
                    annotations[param_name] = param.annotation or "typing.Any"
        event_cls_qualname = f"{cls.__qualname__}.{name}"
        event_cls_dict = {
            "__annotations__": annotations,
            "__module__": cls.__module__,
            "__qualname__": event_cls_qualname,
        }
        if event_topic:
            event_cls_dict["TOPIC"] = event_topic

        # Create the event class object.
        _new_class = type(name, bases, event_cls_dict)
        return cast("type[CanMutateAggregate[Any]]", _new_class)

    def __call__(
        cls: MetaAggregate[TAggregate], *args: Any, **kwargs: Any
    ) -> TAggregate:
        if cls is BaseAggregate:
            msg = "Please define or use subclasses of BaseAggregate."
            raise TypeError(msg)
        created_event_classes = _created_event_classes[cls]
        # Here, unlike when calling _create(), we don't have a given event class,
        # so we need to check that there is one "created" event class to use here.
        # We don't check this in __init_subclass__ to allow for alternatives that
        # can be selected by developers by calling _create(event_class=...).
        if len(created_event_classes) == 0:
            msg = f"No \"created\" event classes defined on class '{cls.__name__}'."
            raise TypeError(msg)

        if len(created_event_classes) > 1:
            msg = (
                f"{cls.__qualname__} can't decide which of many "
                '"created" event classes to use: '
                f"""'{"', '".join(c.__name__ for c in created_event_classes)}'. """
                "Please use class arg 'created_event_name' or"
                " @event decorator on __init__ method."
            )
            raise TypeError(msg)

        kwargs = _coerce_args_to_kwargs(
            cls.__init__,  # type: ignore[misc]
            args,
            kwargs,
            expects_id=cls in _annotations_mention_id,
        )
        return cls._create(
            event_class=created_event_classes[0],
            **kwargs,
        )

    def _create(
        cls: MetaAggregate[TAggregate],
        event_class: type[CanInitAggregate[Any]],
        **kwargs: Any,
    ) -> TAggregate:
        # Just define method signature for the __call__() method.
        raise NotImplementedError  # pragma: no cover


class BaseAggregate(Generic[TAggregateID], metaclass=MetaAggregate):
    """Base class for aggregates."""

    INITIAL_VERSION: int = 1

    @staticmethod
    def create_id(*_: Any, **__: Any) -> TAggregateID:
        """Returns a new aggregate ID."""
        raise NotImplementedError

    @classmethod
    def _create(
        cls: type[Self],
        event_class: type[CanInitAggregate[TAggregateID]],
        *,
        id: TAggregateID | None = None,  # noqa: A002
        **kwargs: Any,
    ) -> Self:
        """Constructs a new aggregate object instance."""
        if getattr(cls, "TOPIC", None):
            _check_explicit_topic_is_registered(event_class)

        # Construct the domain event with an ID and a
        # version, and a topic for the aggregate class.
        create_id_kwargs = {
            k: v for k, v in kwargs.items() if k in _create_id_param_names[cls]
        }
        if id is not None:
            originator_id = id
            if not isinstance(originator_id, (UUID, str)):
                msg = f"Given id was not a UUID or str: {originator_id!r}"
                raise TypeError(msg)
        else:
            try:
                originator_id = cls.create_id(**create_id_kwargs)
            except NotImplementedError as e:
                msg = f"Please pass an 'id' arg or define a create_id() method on {cls}"
                raise NotImplementedError(msg) from e

            if not isinstance(originator_id, (UUID, str)):
                msg = (
                    f"{cls.create_id.__module__}.{cls.create_id.__qualname__}"
                    f" did not return UUID or str, it returned: {originator_id!r}"
                )
                raise TypeError(msg)

        # Impose the required common "created" event attribute values.
        kwargs = kwargs.copy()
        kwargs.update(
            originator_topic=get_topic(cls),
            originator_id=originator_id,
            originator_version=cls.INITIAL_VERSION,
        )
        if kwargs.get("timestamp") is None:
            kwargs["timestamp"] = event_class.create_timestamp()

        try:
            created_event = event_class(**kwargs)
        except TypeError as e:
            msg = f"Unable to construct '{event_class.__qualname__}' event: {e}"
            raise TypeError(msg) from e
        # Construct the aggregate object.
        agg = cast("Self", created_event.mutate(None))

        assert agg is not None
        # Append the domain event to pending list.
        agg.pending_events.append(created_event)
        # Return the aggregate.
        return agg

    def __base_init__(
        self,
        originator_id: Any,
        originator_version: int,
        timestamp: datetime,
    ) -> None:
        """Initialises an aggregate object with an :data:`id`, a :data:`version`
        number, and a :data:`timestamp`.
        """
        self._id: TAggregateID = originator_id
        self._version = originator_version
        self._created_on = timestamp
        self._modified_on = timestamp
        self._pending_events: list[CanMutateAggregate[TAggregateID]] = []

    @property
    def id(self) -> TAggregateID:
        """The ID of the aggregate."""
        return self._id

    @property
    def version(self) -> int:
        """The version number of the aggregate."""
        return self._version

    @version.setter
    def version(self, version: int) -> None:
        self._version = version

    @property
    def created_on(self) -> datetime:
        """The date and time when the aggregate was created."""
        return self._created_on

    @property
    def modified_on(self) -> datetime:
        """The date and time when the aggregate was last modified."""
        return self._modified_on

    @modified_on.setter
    def modified_on(self, modified_on: datetime) -> None:
        self._modified_on = modified_on

    @property
    def pending_events(self) -> list[CanMutateAggregate[TAggregateID]]:
        """A list of pending events."""
        return self._pending_events

    def trigger_event(
        self,
        event_class: type[CanMutateAggregate[TAggregateID]],
        **kwargs: Any,
    ) -> None:
        """Triggers domain event of given type, by creating
        an event object and using it to mutate the aggregate.
        """
        if getattr(type(self), "TOPIC", None):
            if event_class.__name__ == "Event":
                msg = "Triggering base 'Event' class is prohibited."
                raise ProgrammingError(msg)
            _check_explicit_topic_is_registered(event_class)

        # Construct the domain event as the
        # next in the aggregate's sequence.
        # Use counting to generate the sequence.
        next_version = self.version + 1

        # Impose the required common domain event attribute values.
        kwargs = kwargs.copy()
        kwargs.update(
            originator_id=self.id,
            originator_version=next_version,
        )
        if kwargs.get("timestamp") is None:
            kwargs["timestamp"] = event_class.create_timestamp()

        try:
            new_event = event_class(**kwargs)
        except TypeError as e:
            msg = f"Can't construct event {event_class}: {e}"
            raise TypeError(msg) from None

        # Mutate aggregate with domain event.
        new_event.mutate(self)
        # Append the domain event to pending list.
        self._pending_events.append(new_event)

    def collect_events(self) -> Sequence[CanMutateAggregate[TAggregateID]]:
        """Collects and returns a list of pending aggregate
        :class:`AggregateEvent` objects.
        """
        collected = []
        while self._pending_events:
            collected.append(self._pending_events.pop(0))
        return collected

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) and self.__dict__ == other.__dict__

    def __repr__(self) -> str:
        attrs = [
            f"{k.lstrip('_')}={v!r}"
            for k, v in self.__dict__.items()
            if k != "_pending_events"
        ]
        return f"{type(self).__name__}({', '.join(attrs)})"

    def __init_subclass__(
        cls: type[BaseAggregate[TAggregateID]], *, created_event_name: str = ""
    ) -> None:
        """
        Initialises aggregate subclass by defining __init__ method and event classes.
        """
        super().__init_subclass__()

        # Ensure we aren't defining another instance of the same class,
        # because annotations can get confused when using singledispatchmethod
        # during class definition e.g. on an aggregate projector function.
        _module = importlib.import_module(cls.__module__)
        if (
            cls.__name__ in _module.__dict__
            and ENVVAR_DISABLE_REDEFINITION_CHECK not in os.environ
        ):
            msg = (
                f"Name '{cls.__name__}' of {cls} already defined in "
                f"'{cls.__module__}' module: {_module.__dict__[cls.__name__]}"
            )
            raise ProgrammingError(msg)

        # Get the class annotations.
        class_annotations = cls.__dict__.get("__annotations__", {})
        try:
            class_annotations.pop("id")
            _annotations_mention_id.add(cls)
        except KeyError:
            pass

        if "id" in cls.__dict__:
            msg = f"Setting attribute 'id' on class '{cls.__name__}' is not allowed"
            raise ProgrammingError(msg)

        # Process the class as a dataclass, if there are annotations.
        if (
            class_annotations
            or cls in _annotations_mention_id
            or any(dataclasses.is_dataclass(base) for base in cls.__bases__)
        ):
            dataclasses.dataclass(eq=False, repr=False)(cls)

        # Remember if __init__ mentions ID.
        for param_name in inspect.signature(cls.__init__).parameters:
            if param_name == "id":
                _init_mentions_id.add(cls)
                break

        # Analyse __init__ attribute, to get __init__ method and @event decorator.
        init_attr: FunctionType | CommandMethodDecorator | None = cls.__dict__.get(
            "__init__"
        )
        if init_attr is None:
            # No method, no decorator.
            init_method: CallableType | None = None
            init_decorator: CommandMethodDecorator | None = None
        elif isinstance(init_attr, CommandMethodDecorator):
            # Method decorated with @event.
            init_method = init_attr.decorated_func
            init_decorator = init_attr
        else:
            # Undecorated __init__ method.
            init_decorator = None
            init_method = init_attr

        # Identify or define a base event class for this aggregate.
        base_event_name = "Event"
        base_event_cls: type[CanMutateAggregate[TAggregateID]] | None = None
        msg = f"Base event class 'Event' not defined on {cls} or ancestors"
        base_event_class_not_defined_error = TypeError(msg)

        try:
            base_event_cls = cls.__dict__[base_event_name]
        except KeyError:
            try:
                super_base_event_cls = getattr(cls, base_event_name)
            except AttributeError:
                pass
            else:
                base_event_cls = cls._define_event_class(
                    name=base_event_name,
                    bases=(super_base_event_cls,),
                    apply_method=None,
                )
                setattr(cls, base_event_name, base_event_cls)

        # Remember which events have been redefined, to preserve apparent hierarchy,
        # in a mapping from the original class to the redefined class.
        redefined_event_classes: dict[
            type[CanMutateAggregate[TAggregateID]],
            type[CanMutateAggregate[TAggregateID]],
        ] = {}

        # Remember any "created" event classes that are discovered.
        created_event_classes: dict[str, type[CanInitAggregate[TAggregateID]]] = {}

        # TODO: Review decorator processing below to see if subclassing can be improved.
        #  - basically, look at the decorators first, build a plan for defining events

        # Ensure events defined on this class are subclasses of the base event class.
        for name, value in tuple(cls.__dict__.items()):
            # Don't subclass the base event class again.
            if name == base_event_name:
                continue

            # Don't subclass lowercase named attributes.
            if name.lower() == name:
                continue

            # Don't subclass if not "CanMutateAggregate".
            if not isinstance(value, type) or not issubclass(value, CanMutateAggregate):
                continue

            # # Don't subclass generic classes (we don't have a type argument).
            # # TODO: Maybe also prohibit triggering such things?
            # if value.__dict__.get("__parameters__", ()):
            #     continue

            # Check we have a base event class.
            if base_event_cls is None:
                raise base_event_class_not_defined_error

            # Redefine events that aren't already subclass of the base event class.
            if not issubclass(value, base_event_cls):
                # Identify base classes that were redefined, to preserve hierarchy.
                redefined_bases = []
                for base in value.__bases__:
                    if base in redefined_event_classes:
                        redefined_bases.append(redefined_event_classes[base])
                    elif "__pydantic_generic_metadata__" in base.__dict__:
                        pydantic_metadata = base.__dict__[
                            "__pydantic_generic_metadata__"
                        ]
                        for i, key in enumerate(pydantic_metadata):
                            if key == "origin":
                                origin = base.__bases__[i]
                                if origin in redefined_event_classes:
                                    redefined_bases.append(
                                        redefined_event_classes[origin]
                                    )

                # Decide base classes of redefined event class: it must be
                # a subclass of the original class, all redefined classes that
                # were in its bases, and the aggregate's base event class.
                event_class_bases = (
                    value,
                    *redefined_bases,
                    base_event_cls,
                )

                # Define event class.
                event_class = cls._define_event_class(name, event_class_bases, None)
                setattr(cls, name, event_class)

                # Remember which events have been redefined.
                redefined_event_classes[value] = event_class
            else:
                event_class = value

            # Remember all "created" event classes defined on this class.
            if issubclass(event_class, CanInitAggregate):
                created_event_classes[name] = event_class

        # Identify or define the aggregate's "created" event class.
        created_event_class: type[CanInitAggregate[TAggregateID]] | None = None
        created_event_topic: str | None = None

        # Analyse __init__ method decorator.
        if init_decorator:

            # Does the decorator specify an event class?
            if init_decorator.given_event_cls:

                # Disallow conflicts between 'created_event_name' and given class.
                if (
                    created_event_name
                    and created_event_name != init_decorator.given_event_cls.__name__
                ):
                    msg = (
                        "Given 'created_event_name' conflicts "
                        "with decorator on __init__"
                    )
                    raise TypeError(msg)

                # Check given event class can init aggregate.
                if not issubclass(init_decorator.given_event_cls, CanInitAggregate):
                    msg = (
                        f"class '{init_decorator.given_event_cls.__name__}' "
                        f'not a "created" event class'
                    )
                    raise TypeError(msg)

                # Have we already subclassed the given event class?
                for sub_class in created_event_classes.values():
                    if issubclass(sub_class, init_decorator.given_event_cls):
                        created_event_class = sub_class
                        break
                else:
                    created_event_class = init_decorator.given_event_cls

            # Does the decorator specify an event name?
            elif init_decorator.event_cls_name:
                created_event_topic = init_decorator.event_topic
                # Disallow conflicts between 'created_event_name' and given name.
                if (
                    created_event_name
                    and created_event_name != init_decorator.event_cls_name
                ):
                    msg = (
                        "Given 'created_event_name' conflicts "
                        "with decorator on __init__"
                    )
                    raise TypeError(msg)

                created_event_name = init_decorator.event_cls_name

            # Disallow using decorator on __init__ without event name or class.
            else:
                msg = "@event decorator on __init__ has neither event name nor class"
                raise TypeError(msg)

        # Do we need to define a created event class?
        if not created_event_class:
            # If we have a "created" event class that matches the name, then use it.
            if created_event_name in created_event_classes:
                created_event_class = created_event_classes[created_event_name]
            # Otherwise, if we have no name and only one class defined, then use it.
            elif not created_event_name and len(created_event_classes) == 1:
                created_event_class = next(iter(created_event_classes.values()))

            # Otherwise, if there are no "created" event classes, or a name
            # is specified that hasn't matched, then try to define one.
            elif len(created_event_classes) == 0 or created_event_name:
                # Decide the base "created" event class.

                base_created_event_cls: type[CanInitAggregate[TAggregateID]] | None = (
                    None
                )

                if created_event_name:
                    # Look for a base class with the same name.
                    with contextlib.suppress(AttributeError):
                        base_created_event_cls = cast(
                            type[CanInitAggregate[TAggregateID]],
                            getattr(cls, created_event_name),
                        )

                if base_created_event_cls is None:
                    # Look for base class with one nominated "created" event.
                    for base_cls in cls.__mro__:
                        if (
                            base_cls in _created_event_classes
                            and len(_created_event_classes[base_cls]) == 1
                        ):
                            base_created_event_cls = _created_event_classes[base_cls][0]
                            break

                if base_created_event_cls:
                    if not created_event_name:
                        created_event_name = base_created_event_cls.__name__

                    # Disallow init method from having variable params, because
                    # we are using it to define a "created" event class.
                    if init_method:
                        _raise_type_error_if_func_has_variable_params(init_method)

                    # Sanity check: we have a base event class.
                    assert base_event_cls is not None
                    # Sanity check: the base created event class is a class.
                    assert isinstance(
                        base_created_event_cls, type
                    ), base_created_event_cls
                    # Sanity check: base created event not subclass of base event class.
                    assert not issubclass(
                        base_created_event_cls, base_event_cls
                    ), base_created_event_cls

                    # Define "created" event class.
                    assert created_event_name
                    assert issubclass(base_created_event_cls, CanInitAggregate)
                    created_event_class_bases = (base_created_event_cls, base_event_cls)
                    created_event_class = cast(
                        type[CanInitAggregate[TAggregateID]],
                        cls._define_event_class(
                            created_event_name,
                            created_event_class_bases,
                            init_method,
                            event_topic=created_event_topic,
                        ),
                    )
                    # Set the event class as an attribute of the aggregate class.
                    setattr(cls, created_event_name, created_event_class)

                elif created_event_name:
                    msg = (
                        'Can\'t defined "created" event class '
                        f"for name '{created_event_name}'"
                    )
                    raise TypeError(msg)

        if created_event_class:
            _created_event_classes[cls] = [created_event_class]
        else:
            # Prepare to disallow any ambiguity of choice between created event classes.
            _created_event_classes[cls] = list(created_event_classes.values())

        # Find and analyse any @event decorators.
        for attr_name, attr_value in tuple(cls.__dict__.items()):
            event_decorator: CommandMethodDecorator | None = None

            # Ignore a decorator on the __init__ method.
            if isinstance(attr_value, CommandMethodDecorator) and (
                attr_value.decorated_func.__name__ == "__init__"
            ):
                continue

            # Handle @property.setter decorator on top of @event decorator.
            if isinstance(attr_value, property) and isinstance(
                attr_value.fset, CommandMethodDecorator
            ):
                event_decorator = attr_value.fset
                # Inspect the setter method.
                method_signature = inspect.signature(event_decorator.decorated_func)
                assert len(method_signature.parameters) == 2
                event_decorator.is_property_setter = True
                event_decorator.property_setter_arg_name = list(
                    method_signature.parameters
                )[1]
                if event_decorator.decorated_func.__name__ != attr_name:
                    attr = cls.__dict__[event_decorator.decorated_func.__name__]
                    if isinstance(attr, CommandMethodDecorator):
                        # This is the "x = property(getx, setx) form" where setx
                        # is a decorated method.
                        continue
                        # Otherwise, it's "x = property(getx, event(setx))".
                elif event_decorator.is_name_inferred_from_method:
                    # This is the "@property.setter \ @event" form. We don't want
                    # event class name inferred from property (not past participle).
                    method_name = event_decorator.decorated_func.__name__
                    msg = (
                        f"@event decorator under @{method_name}.setter "
                        "requires event name or class"
                    )
                    raise TypeError(msg)

            elif isinstance(attr_value, CommandMethodDecorator):
                event_decorator = attr_value

            if event_decorator is not None:
                if event_decorator.given_event_cls:
                    # Check this is not a "created" event class.
                    if issubclass(event_decorator.given_event_cls, CanInitAggregate):
                        msg = (
                            f"{event_decorator.given_event_cls} "
                            f"is subclass of {CanInitAggregate.__name__}"
                        )
                        raise TypeError(msg)

                    # Define event class as subclass of given class.
                    given_subclass = cast(
                        type[CanMutateAggregate[TAggregateID]],
                        getattr(cls, event_decorator.given_event_cls.__name__),
                    )
                    # TODO: Check if this subclassing means we can avoid some of
                    #  the subclassing of events above? Maybe do this first?
                    event_cls = cls._define_event_class(
                        event_decorator.given_event_cls.__name__,
                        (DecoratedFuncCaller, given_subclass),
                        None,
                    )

                else:
                    # Check event class isn't already defined.
                    assert event_decorator.event_cls_name
                    if event_decorator.event_cls_name in cls.__dict__:
                        msg = (
                            f"{event_decorator.event_cls_name} "
                            f"event already defined on {cls.__name__}"
                        )
                        raise TypeError(msg)

                    # Check we have a base event class.
                    if base_event_cls is None:
                        raise base_event_class_not_defined_error

                    # Define event class from signature of original method.
                    event_cls = cls._define_event_class(
                        event_decorator.event_cls_name,
                        (DecoratedFuncCaller, base_event_cls),
                        event_decorator.decorated_func,
                        event_topic=event_decorator.event_topic,
                    )

                # Cache the decorated method for the event class to use.
                decorated_funcs[event_cls] = event_decorator.decorated_func

                # Set the event class as an attribute of the aggregate class.
                setattr(cls, event_cls.__name__, event_cls)

                # Remember which event class to trigger.
                decorated_func_callers[event_decorator] = cast(
                    type[DecoratedFuncCaller], event_cls
                )

        # Check any create_id() method defined on this class is static or class method.
        if "create_id" in cls.__dict__ and not isinstance(
            cls.__dict__["create_id"], (staticmethod, classmethod)
        ):
            msg = (
                f"{cls.create_id} is not a static or class method: "
                f"{type(cls.create_id)}"
            )
            raise TypeError(msg)

        # Get the parameters of the create_id method that will be used by this class.
        for name, param in inspect.signature(cls.create_id).parameters.items():
            if param.kind in [param.KEYWORD_ONLY, param.POSITIONAL_OR_KEYWORD]:
                _create_id_param_names[cls].append(name)

        # Define event classes for all events on all bases if not defined on this class.
        for aggregate_base_class in cls.__bases__:
            for name, value in aggregate_base_class.__dict__.items():
                if (
                    isinstance(value, type)
                    and issubclass(value, CanMutateAggregate)
                    and name not in cls.__dict__
                    and name.lower() != name
                ):
                    # Sanity check: we have a base event class.
                    assert base_event_cls is not None
                    event_class = cls._define_event_class(
                        name, (base_event_cls, value), None
                    )
                    setattr(cls, name, event_class)

        if getattr(cls, "TOPIC", None):

            explicit_topic = cls.__dict__.get("TOPIC", None)

            if not explicit_topic:
                msg = f"Explicit topic not defined on {cls}"
                raise ProgrammingError(msg)

            try:
                register_topic(explicit_topic, cls)
            except TopicError:
                msg = (
                    f"Explicit topic '{explicit_topic}' of {cls} "
                    f"already registered for {resolve_topic(explicit_topic)}"
                )
                raise ProgrammingError(msg) from None

            for name, obj in cls.__dict__.items():
                if (
                    isinstance(obj, type)
                    and issubclass(obj, CanMutateAggregate)
                    and name != "Event"
                ):
                    explicit_topic = getattr(obj, "TOPIC", None)
                    if not explicit_topic:
                        msg = f"Explicit topic not defined on {obj}"
                        raise ProgrammingError(msg)
                    try:
                        register_topic(explicit_topic, obj)
                    except TopicError:
                        msg = (
                            f"Explicit topic '{explicit_topic}' of {obj} "
                            f"already registered for {resolve_topic(explicit_topic)}"
                        )
                        raise ProgrammingError(msg) from None

    def __hash__(self) -> int:
        raise NotImplementedError  # pragma: no cover


def _check_explicit_topic_is_registered(event_class: type[object]) -> None:
    explicit_topic = getattr(event_class, "TOPIC", None)
    if not explicit_topic:
        msg = f"Explicit topic not defined on {event_class}"
        raise ProgrammingError(msg)
    try:
        resolved_obj = resolve_topic(explicit_topic)
    except TopicError:
        msg = f"Explicit topic '{explicit_topic}' on {event_class} is not registered"
        raise ProgrammingError(msg) from None
    if resolved_obj is not event_class:
        msg = (
            f"Explicit topic '{explicit_topic}' on {event_class} "
            f"already registered for {resolved_obj}"
        )
        raise ProgrammingError(msg) from None


class OriginatorIDError(EventSourcingError):
    """Raised when a domain event can't be applied to
    an aggregate due to an ID mismatch indicating
    the domain event is not in the aggregate's
    sequence of events.
    """


class OriginatorVersionError(EventSourcingError):
    """Raised when a domain event can't be applied to
    an aggregate due to version mismatch indicating
    the domain event is not the next in the aggregate's
    sequence of events.
    """


class SnapshotProtocol(DomainEventProtocol[TAggregateID_co], Protocol):
    @property
    def state(self) -> Any:
        """Snapshots have a read-only 'state'."""
        raise NotImplementedError  # pragma: no cover

    # TODO: Improve on this 'Any'.
    @classmethod
    def take(cls: Any, aggregate: Any) -> Any:
        """Snapshots have a 'take()' class method."""


class CanSnapshotAggregate(HasOriginatorIDVersion[TAggregateID], CanCreateTimestamp):
    topic: str
    state: Any

    def __init_subclass__(cls) -> None:
        cls.find_originator_id_type(CanSnapshotAggregate)
        super().__init_subclass__()

    # def __init__(
    #     self,
    #     originator_id: UUID,
    #     originator_version: int,
    #     timestamp: datetime,
    #     topic: str,
    #     state: Any,
    # ) -> None:
    #     raise NotImplementedError  # pragma: no cover

    @classmethod
    def take(
        cls,
        aggregate: MutableOrImmutableAggregate[TAggregateID],
    ) -> Self:
        """Creates a snapshot of the given :class:`Aggregate` object."""
        aggregate_state = dict(vars(aggregate))
        class_version = getattr(type(aggregate), "class_version", 1)
        if class_version > 1:
            aggregate_state["class_version"] = class_version
        if isinstance(aggregate, Aggregate):
            aggregate_state.pop("_id")
            aggregate_state.pop("_version")
            aggregate_state.pop("_pending_events")
        return cls(
            originator_id=aggregate.id,  # type: ignore[call-arg]
            originator_version=aggregate.version,  # pyright: ignore[reportCallIssue]
            timestamp=cls.create_timestamp(),  # pyright: ignore[reportCallIssue]
            topic=get_topic(type(aggregate)),  # pyright: ignore[reportCallIssue]
            state=aggregate_state,  # pyright: ignore[reportCallIssue]
        )

    def mutate(self, _: None) -> BaseAggregate[TAggregateID]:
        """Reconstructs the snapshotted :class:`Aggregate` object."""
        cls = cast(type[BaseAggregate[TAggregateID]], resolve_topic(self.topic))
        aggregate_state = dict(self.state)
        from_version = aggregate_state.pop("class_version", 1)
        class_version = getattr(cls, "class_version", 1)
        while from_version < class_version:
            upcast_name = f"upcast_v{from_version}_v{from_version + 1}"
            upcast = getattr(cls, upcast_name)
            upcast(aggregate_state)
            from_version += 1

        aggregate_state["_id"] = self.originator_id
        aggregate_state["_version"] = self.originator_version
        aggregate_state["_pending_events"] = []
        aggregate = object.__new__(cls)
        object.__setattr__(aggregate, "__dict__", aggregate_state)
        return aggregate


@dataclass(frozen=True)
class Snapshot(CanSnapshotAggregate[UUID], DomainEvent):
    """Snapshots represent the state of an aggregate at a particular
    version.

    Constructor arguments:

    :param UUID originator_id: ID of originating aggregate.
    :param int originator_version: version of originating aggregate.
    :param datetime timestamp: date-time of the event
    :param str topic: string that includes a class and its module
    :param dict state: state of originating aggregate.
    """

    topic: str
    state: dict[str, Any]


class Aggregate(BaseAggregate[UUID]):
    @staticmethod
    def create_id(*_: Any, **__: Any) -> UUID:
        """Returns a new aggregate ID."""
        return uuid4()

    class Event(AggregateEvent):
        pass

    class Created(Event, AggregateCreated):
        pass

    Snapshot = Snapshot


@overload
def aggregate(*, created_event_name: str) -> Callable[[Any], type[Aggregate]]:
    pass  # pragma: no cover


@overload
def aggregate(cls: Any) -> type[Aggregate]:
    pass  # pragma: no cover


def aggregate(
    cls: Any | None = None,
    *,
    created_event_name: str = "",
) -> type[Aggregate] | Callable[[Any], type[Aggregate]]:
    """Converts the class that was passed in to inherit from Aggregate.

    .. code-block:: python

        @aggregate
        class MyAggregate:
            pass

    ...is equivalent to...

    .. code-block:: python

        class MyAggregate(Aggregate):
            pass
    """

    def decorator(cls_: Any) -> type[Aggregate]:
        if issubclass(cls_, Aggregate):
            msg = f"{cls_.__qualname__} is already an Aggregate"
            raise TypeError(msg)
        bases = cls_.__bases__
        if bases == (object,):
            bases = (Aggregate,)
        else:
            bases += (Aggregate,)
        cls_dict = {}
        cls_dict.update(cls_.__dict__)
        cls_ = MetaAggregate(
            cls_.__qualname__,
            bases,
            cls_dict,
            created_event_name=created_event_name,
        )
        assert issubclass(cls_, Aggregate)
        return cls_

    if cls:
        return decorator(cls)
    return decorator



from __future__ import annotations

import zlib

from eventsourcing.persistence import Compressor


class ZlibCompressor(Compressor):
    def compress(self, data: bytes) -> bytes:
        """Compress bytes using zlib."""
        return zlib.compress(data)

    def decompress(self, data: bytes) -> bytes:
        """Decompress bytes using zlib."""
        return zlib.decompress(data)



from __future__ import annotations

import inspect
import threading
import traceback
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Sequence
from queue import Full, Queue
from types import FrameType, ModuleType
from typing import TYPE_CHECKING, Any, ClassVar, Generic, cast

from eventsourcing.application import (
    Application,
    NotificationLog,
    ProgrammingError,
    Section,
    TApplication,
)
from eventsourcing.domain import (
    DomainEventProtocol,
    MutableOrImmutableAggregate,
    TAggregateID,
)
from eventsourcing.persistence import (
    Mapper,
    Notification,
    ProcessRecorder,
    Recording,
    Tracking,
)
from eventsourcing.projection import EventSourcedProjection
from eventsourcing.utils import EnvType, get_topic, resolve_topic

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

    from typing_extensions import Self


ProcessingJob = tuple[DomainEventProtocol[TAggregateID], Tracking]


class RecordingEvent(Generic[TAggregateID]):
    def __init__(
        self,
        application_name: str,
        recordings: list[Recording[TAggregateID]],
        previous_max_notification_id: int | None,
    ):
        self.application_name = application_name
        self.recordings: list[Recording[TAggregateID]] = recordings
        self.previous_max_notification_id = previous_max_notification_id


ConvertingJob = RecordingEvent[TAggregateID] | Sequence[Notification] | None


class Follower(EventSourcedProjection[TAggregateID]):
    """Extends the :class:`~eventsourcing.projection.EventSourcedProjection` class
    by pulling notification objects from its notification log readers, by converting
    the notification objects to domain events and tracking objects and by processing
    the reconstructed domain event objects.
    """

    pull_section_size = 10

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        # for backwards compatibility, set "topics" if has "follow_topics".
        cls.topics = getattr(cls, "follow_topics", cls.topics)

    def __init__(self, env: EnvType | None = None) -> None:
        super().__init__(env)
        self.readers: dict[str, NotificationLogReader] = {}
        self.mappers: dict[str, Mapper[TAggregateID]] = {}
        self.is_threading_enabled = False

    def follow(self, name: str, log: NotificationLog) -> None:
        """Constructs a notification log reader and a mapper for
        the named application, and adds them to its collections
        of readers and mappers.
        """
        assert isinstance(self.recorder, ProcessRecorder)
        reader = NotificationLogReader(log, section_size=self.pull_section_size)
        env = self.construct_env(name, self.env)
        factory = self.construct_factory(env)
        mapper = factory.mapper(
            self.construct_transcoder(), mapper_class=type(self.mapper)
        )
        self.readers[name] = reader
        self.mappers[name] = mapper

    # @retry(IntegrityError, max_attempts=100)
    def pull_and_process(
        self, leader_name: str, start: int | None = None, stop: int | None = None
    ) -> None:
        """Pull and process new domain event notifications."""
        if start is None:
            start = self.recorder.max_tracking_id(leader_name)
        for notifications in self.pull_notifications(
            leader_name, start=start, stop=stop, inclusive_of_start=False
        ):
            notifications_iter = self.filter_received_notifications(notifications)
            for domain_event, tracking in self.convert_notifications(
                leader_name, notifications_iter
            ):
                self.process_event(domain_event, tracking)

    def process_event(
        self, domain_event: DomainEventProtocol[TAggregateID], tracking: Tracking
    ) -> None:
        with self.processing_lock:
            super().process_event(domain_event, tracking)

    def pull_notifications(
        self,
        leader_name: str,
        start: int | None,
        stop: int | None = None,
        *,
        inclusive_of_start: bool = True,
    ) -> Iterator[Sequence[Notification]]:
        """Pulls batches of unseen :class:`~eventsourcing.persistence.Notification`
        objects from the notification log reader of the named application.
        """
        return self.readers[leader_name].select(
            start=start,
            stop=stop,
            topics=self.topics,
            inclusive_of_start=inclusive_of_start,
        )

    def filter_received_notifications(
        self, notifications: Sequence[Notification]
    ) -> Sequence[Notification]:
        if self.topics:
            return [n for n in notifications if n.topic in self.topics]
        return notifications

    def convert_notifications(
        self, leader_name: str, notifications: Iterable[Notification]
    ) -> list[ProcessingJob[TAggregateID]]:
        """Uses the given :class:`~eventsourcing.persistence.Mapper` to convert
        each received :class:`~eventsourcing.persistence.Notification`
        object to an :class:`~eventsourcing.domain.AggregateEvent` object
        paired with a :class:`~eventsourcing.persistence.Tracking` object.
        """
        mapper = self.mappers[leader_name]
        processing_jobs = []
        for notification in notifications:
            domain_event: DomainEventProtocol[TAggregateID] = mapper.to_domain_event(
                notification
            )
            tracking = Tracking(
                application_name=leader_name,
                notification_id=notification.id,
            )
            processing_jobs.append((domain_event, tracking))
        return processing_jobs


class RecordingEventReceiver(ABC, Generic[TAggregateID]):
    """Abstract base class for objects that may receive recording events."""

    @abstractmethod
    def receive_recording_event(
        self, new_recording_event: RecordingEvent[TAggregateID]
    ) -> None:
        """Receives a recording event."""


class Leader(Application[TAggregateID]):
    """Extends the :class:`~eventsourcing.application.Application`
    class by also being responsible for keeping track of
    followers, and prompting followers when there are new
    domain event notifications to be pulled and processed.
    """

    def __init__(self, env: EnvType | None = None) -> None:
        super().__init__(env)
        self.previous_max_notification_id: int | None = None
        self.followers: list[RecordingEventReceiver[TAggregateID]] = []

    def lead(self, follower: RecordingEventReceiver[TAggregateID]) -> None:
        """Adds given follower to a list of followers."""
        self.followers.append(follower)

    def save(
        self,
        *objs: MutableOrImmutableAggregate[TAggregateID]
        | DomainEventProtocol[TAggregateID]
        | None,
        **kwargs: Any,
    ) -> list[Recording[TAggregateID]]:
        if self.previous_max_notification_id is None:
            self.previous_max_notification_id = self.recorder.max_notification_id()
        return super().save(*objs, **kwargs)

    def _notify(self, recordings: list[Recording[TAggregateID]]) -> None:
        """Calls :func:`receive_recording_event` on each follower
        whenever new events have just been saved.
        """
        super()._notify(recordings)
        if self.notify_topics:
            recordings = [
                r for r in recordings if r.notification.topic in self.notify_topics
            ]
        if recordings:
            recording_event = RecordingEvent(
                application_name=self.name,
                recordings=recordings,
                previous_max_notification_id=self.previous_max_notification_id,
            )
            self.previous_max_notification_id = recordings[-1].notification.id
            for follower in self.followers:
                follower.receive_recording_event(recording_event)


class ProcessApplication(Leader[TAggregateID], Follower[TAggregateID]):
    """Base class for event processing applications
    that are both "leaders" and followers".
    """


class System:
    """Defines a system of applications."""

    __caller_modules: ClassVar[dict[int, ModuleType]] = {}

    def __init__(
        self,
        pipes: Iterable[Iterable[type[Application[Any]]]],
    ):
        # Remember the caller frame's module, so that we might identify a topic.
        caller_frame = cast(FrameType, inspect.currentframe()).f_back
        module = cast(ModuleType, inspect.getmodule(caller_frame))
        type(self).__caller_modules[id(self)] = module  # noqa: SLF001

        # Build nodes and edges.
        self.edges: list[tuple[str, str]] = []
        classes: dict[str, type[Application[Any]]] = {}
        for pipe in pipes:
            follower_cls = None
            for cls in pipe:
                classes[cls.name] = cls
                if follower_cls is None:
                    follower_cls = cls
                else:
                    leader_cls = follower_cls
                    follower_cls = cls
                    edge = (leader_cls.name, follower_cls.name)
                    if edge not in self.edges:
                        self.edges.append(edge)

        self.nodes: dict[str, str] = {}
        for name, cls in classes.items():
            self.nodes[name] = get_topic(cls)

        # Identify leaders and followers.
        self.follows: dict[str, list[str]] = defaultdict(list)
        self.leads: dict[str, list[str]] = defaultdict(list)
        for edge in self.edges:
            self.leads[edge[0]].append(edge[1])
            self.follows[edge[1]].append(edge[0])

        # Identify singles.
        self.singles = []
        for name in classes:
            if name not in self.leads and name not in self.follows:
                self.singles.append(name)

        # Check followers are followers.
        for name in self.follows:
            if not issubclass(classes[name], Follower):
                msg = f"Not a follower class: {classes[name]}"
                raise TypeError(msg)

        # Check each process is a process application class.
        for name in self.processors:
            if not issubclass(classes[name], ProcessApplication):
                msg = f"Not a process application class: {classes[name]}"
                raise TypeError(msg)

    @property
    def leaders(self) -> list[str]:
        return list(self.leads.keys())

    @property
    def leaders_only(self) -> list[str]:
        return [name for name in self.leads if name not in self.follows]

    @property
    def followers(self) -> list[str]:
        return list(self.follows.keys())

    @property
    def processors(self) -> list[str]:
        return [name for name in self.leads if name in self.follows]

    def get_app_cls(self, name: str) -> type[Application[Any]]:
        cls = resolve_topic(self.nodes[name])
        assert issubclass(cls, Application)
        return cls

    def leader_cls(self, name: str) -> type[Leader[Any]]:
        cls = self.get_app_cls(name)
        if issubclass(cls, Leader):
            return cls
        cls = type(cls.name, (Leader, cls), {})
        assert issubclass(cls, Leader)
        return cls

    def follower_cls(self, name: str) -> type[Follower[Any]]:
        cls = self.get_app_cls(name)
        assert issubclass(cls, Follower)
        return cls

    @property
    def topic(self) -> str:
        """
        Returns a topic to the system object, if constructed as a module attribute.
        """
        topic: str | None = None
        module = System.__caller_modules[id(self)]
        for name, value in module.__dict__.items():
            if value is self:
                topic = module.__name__ + ":" + name
                assert resolve_topic(topic) is self
        if topic is None:
            msg = f"Unable to compute topic for system object: {self}"
            raise ProgrammingError(msg)
        return topic


class Runner(ABC, Generic[TAggregateID]):
    """Abstract base class for system runners."""

    def __init__(self, system: System, env: EnvType | None = None):
        self.system = system
        self.env = env
        self.is_started = False

    @abstractmethod
    def start(self) -> None:
        """Starts the runner."""
        if self.is_started:
            raise RunnerAlreadyStartedError
        self.is_started = True

    @abstractmethod
    def stop(self) -> None:
        """Stops the runner."""

    @abstractmethod
    def get(self, cls: type[TApplication]) -> TApplication:
        """Returns an application instance for given application class."""

    def __enter__(self) -> Self:
        self.start()
        return self

    def __exit__(self, *args: object, **kwargs: Any) -> None:
        self.stop()


class RunnerAlreadyStartedError(Exception):
    """Raised when runner is already started."""


class NotificationPullingError(Exception):
    """Raised when pulling notifications fails."""


class NotificationConvertingError(Exception):
    """Raised when converting notifications fails."""


class EventProcessingError(Exception):
    """Raised when event processing fails."""


class SingleThreadedRunner(Runner[TAggregateID], RecordingEventReceiver[TAggregateID]):
    """Runs a :class:`System` in a single thread."""

    def __init__(self, system: System, env: EnvType | None = None):
        """Initialises runner with the given :class:`System`."""
        super().__init__(system=system, env=env)
        self.apps: dict[str, Application[TAggregateID]] = {}
        self._recording_events_received: list[RecordingEvent[TAggregateID]] = []
        self._prompted_names_lock = threading.Lock()
        self._prompted_names: set[str] = set()
        self._processing_lock = threading.Lock()

        # Construct followers.
        for name in self.system.followers:
            self.apps[name] = self.system.follower_cls(name)(env=self.env)

        # Construct leaders.
        for name in self.system.leaders_only:
            leader = self.system.leader_cls(name)(env=self.env)
            self.apps[name] = leader

        # Construct singles.
        for name in self.system.singles:
            single = self.system.get_app_cls(name)(env=self.env)
            self.apps[name] = single

    def start(self) -> None:
        """Starts the runner. The applications mentioned in the system definition
        are constructed. The followers are set up to follow the applications
        they are defined as following in the system definition. And the leaders
        are set up to lead the runner itself.
        """
        super().start()

        # Setup followers to follow leaders.
        for edge in self.system.edges:
            leader_name = edge[0]
            follower_name = edge[1]
            leader = cast("Leader[Any]", self.apps[leader_name])
            follower = cast(Follower[Any], self.apps[follower_name])
            assert isinstance(leader, Leader)
            assert isinstance(follower, Follower)
            follower.follow(leader_name, leader.notification_log)

        # Setup leaders to lead this runner.
        for name in self.system.leaders:
            leader = cast("Leader[Any]", self.apps[name])
            assert isinstance(leader, Leader)
            leader.lead(self)

    def receive_recording_event(
        self, new_recording_event: RecordingEvent[TAggregateID]
    ) -> None:
        """Receives recording event by appending the name of the leader
        to a list of prompted names.

        Then, unless this method has previously been called and not yet returned,
        each of the prompted names is resolved to a leader application, and its
        followers pull and process events from that application. This may lead to
        further names being added to the list of prompted names. This process
        continues until there are no more prompted names. In this way, a system
        of applications will process all events in a single thread.
        """
        leader_name = new_recording_event.application_name
        with self._prompted_names_lock:
            self._prompted_names.add(leader_name)

        if self._processing_lock.acquire(blocking=False):
            try:
                while True:
                    with self._prompted_names_lock:
                        prompted_names = self._prompted_names
                        self._prompted_names = set()

                        if not prompted_names:
                            break

                    for leader_name in prompted_names:
                        for follower_name in self.system.leads[leader_name]:
                            follower = cast(Follower[Any], self.apps[follower_name])
                            follower.pull_and_process(leader_name)

            finally:
                self._processing_lock.release()

    def stop(self) -> None:
        for app in self.apps.values():
            app.close()
        self.apps.clear()

    def get(self, cls: type[TApplication]) -> TApplication:
        app = self.apps[cls.name]
        assert isinstance(app, cls)
        return app


class NewSingleThreadedRunner(
    Runner[TAggregateID], RecordingEventReceiver[TAggregateID]
):
    """Runs a :class:`System` in a single thread."""

    def __init__(self, system: System, env: EnvType | None = None):
        """Initialises runner with the given :class:`System`."""
        super().__init__(system=system, env=env)
        self.apps: dict[str, Application[Any]] = {}
        self._recording_events_received: list[RecordingEvent[TAggregateID]] = []
        self._recording_events_received_lock = threading.Lock()
        self._processing_lock = threading.Lock()
        self._previous_max_notification_ids: dict[str, int] = {}

        # Construct followers.
        for name in self.system.followers:
            self.apps[name] = self.system.follower_cls(name)(env=self.env)

        # Construct leaders.
        for name in self.system.leaders_only:
            leader = self.system.leader_cls(name)(env=self.env)
            self.apps[name] = leader

        # Construct singles.
        for name in self.system.singles:
            single = self.system.get_app_cls(name)(env=self.env)
            self.apps[name] = single

    def start(self) -> None:
        """Starts the runner.
        The applications are constructed, and setup to lead and follow
        each other, according to the system definition.
        The followers are setup to follow the applications they follow
        (have a notification log reader with the notification log of the
        leader), and their leaders are setup to lead the runner itself
        (send prompts).
        """
        super().start()

        # Setup followers to follow leaders.
        for edge in self.system.edges:
            leader_name = edge[0]
            follower_name = edge[1]
            leader = cast("Leader[Any]", self.apps[leader_name])
            follower = cast(Follower[Any], self.apps[follower_name])
            assert isinstance(leader, Leader)
            assert isinstance(follower, Follower)
            follower.follow(leader_name, leader.notification_log)

        # Setup leaders to notify followers.
        for name in self.system.leaders:
            leader = cast("Leader[Any]", self.apps[name])
            assert isinstance(leader, Leader)
            leader.lead(self)

    def receive_recording_event(
        self, new_recording_event: RecordingEvent[TAggregateID]
    ) -> None:
        """Receives recording event by appending it to list of received recording
        events.

        Unless this method has previously been called and not yet returned, it
        will then attempt to make the followers process all received recording
        events, until there are none remaining.
        """
        with self._recording_events_received_lock:
            self._recording_events_received.append(new_recording_event)

        if self._processing_lock.acquire(blocking=False):
            try:
                while True:
                    with self._recording_events_received_lock:
                        recording_events = self._recording_events_received
                        self._recording_events_received = []

                        if not recording_events:
                            break

                    for recording_event in recording_events:
                        leader_name = recording_event.application_name
                        previous_max_notification_id = (
                            self._previous_max_notification_ids.get(leader_name, 0)
                        )

                        # Ignore recording event if already seen a subsequent.
                        if (
                            recording_event.previous_max_notification_id is not None
                            and recording_event.previous_max_notification_id
                            < previous_max_notification_id
                        ):
                            continue

                        # Catch up if there is a gap in sequence of recording events.
                        if (
                            recording_event.previous_max_notification_id is None
                            or recording_event.previous_max_notification_id
                            > previous_max_notification_id
                        ):
                            for follower_name in self.system.leads[leader_name]:
                                follower = self.apps[follower_name]
                                assert isinstance(follower, Follower)
                                start = follower.recorder.max_tracking_id(leader_name)
                                stop = recording_event.recordings[0].notification.id - 1
                                follower.pull_and_process(
                                    leader_name=leader_name,
                                    start=start,
                                    stop=stop,
                                )
                        for recording in recording_event.recordings:
                            for follower_name in self.system.leads[leader_name]:
                                follower = self.apps[follower_name]
                                assert isinstance(follower, Follower)
                                if (
                                    follower.topics
                                    and recording.notification.topic
                                    not in follower.topics
                                ):
                                    continue
                                follower.process_event(
                                    domain_event=recording.domain_event,
                                    tracking=Tracking(
                                        application_name=recording_event.application_name,
                                        notification_id=recording.notification.id,
                                    ),
                                )

                        self._previous_max_notification_ids[leader_name] = (
                            recording_event.recordings[-1].notification.id
                        )

            finally:
                self._processing_lock.release()

    def stop(self) -> None:
        for app in self.apps.values():
            app.close()
        self.apps.clear()

    def get(self, cls: type[TApplication]) -> TApplication:
        app = self.apps[cls.name]
        assert isinstance(app, cls)
        return app


class MultiThreadedRunner(Runner[TAggregateID]):
    """Runs a :class:`System` with one :class:`MultiThreadedRunnerThread`
    for each :class:`Follower` in the system definition.
    """

    def __init__(self, system: System, env: EnvType | None = None):
        """Initialises runner with the given :class:`System`."""
        super().__init__(system=system, env=env)
        self.apps: dict[str, Application[Any]] = {}
        self.threads: dict[str, MultiThreadedRunnerThread[TAggregateID]] = {}
        self.has_errored = threading.Event()

        # Construct followers.
        for follower_name in self.system.followers:
            follower_class = self.system.follower_cls(follower_name)
            try:
                follower = follower_class(env=self.env)
            except Exception:
                self.has_errored.set()
                raise
            self.apps[follower_name] = follower

        # Construct non-follower leaders.
        for leader_name in self.system.leaders_only:
            self.apps[leader_name] = self.system.leader_cls(leader_name)(env=self.env)

        # Construct singles.
        for name in self.system.singles:
            single = self.system.get_app_cls(name)(env=self.env)
            self.apps[name] = single

    def start(self) -> None:
        """Starts the runner.
        A multi-threaded runner thread is started for each
        'follower' application in the system, and constructs
        an instance of each non-follower leader application in
        the system. The followers are then setup to follow the
        applications they follow (have a notification log reader
        with the notification log of the leader), and their leaders
        are  setup to lead the follower's thead (send prompts).
        """
        super().start()

        # Construct followers.
        thread: MultiThreadedRunnerThread[TAggregateID]
        for follower_name in self.system.followers:
            follower = cast(Follower[Any], self.apps[follower_name])

            thread = MultiThreadedRunnerThread(
                follower=follower,
                has_errored=self.has_errored,
            )
            self.threads[follower.name] = thread
            thread.start()

        # Wait until all the threads have started.
        for thread in self.threads.values():
            thread.has_started.wait()

        # Lead and follow.
        for edge in self.system.edges:
            leader = cast("Leader[Any]", self.apps[edge[0]])
            follower = cast(Follower[Any], self.apps[edge[1]])
            follower.follow(leader.name, leader.notification_log)
            thread = self.threads[follower.name]
            leader.lead(thread)

    def watch_for_errors(self, timeout: float | None = None) -> bool:
        if self.has_errored.wait(timeout=timeout):
            self.stop()
        return self.has_errored.is_set()

    def stop(self) -> None:
        threads = self.threads.values()
        for thread in threads:
            thread.stop()
        for thread in threads:
            thread.join(timeout=2)
        for app in self.apps.values():
            app.close()
        self.apps.clear()
        self.reraise_thread_errors()

    def reraise_thread_errors(self) -> None:
        for thread in self.threads.values():
            if thread.error:
                raise thread.error

    def get(self, cls: type[TApplication]) -> TApplication:
        app = self.apps[cls.name]
        assert isinstance(app, cls)
        return app


class MultiThreadedRunnerThread(RecordingEventReceiver[TAggregateID], threading.Thread):
    """Runs one :class:`~eventsourcing.system.Follower` application in
    a :class:`~eventsourcing.system.MultiThreadedRunner`.
    """

    def __init__(
        self,
        follower: Follower[Any],
        has_errored: threading.Event,
    ):
        super().__init__(daemon=True)
        self.follower = follower
        self.has_errored = has_errored
        self.error: Exception | None = None
        self.is_stopping = threading.Event()
        self.has_started = threading.Event()
        self.is_prompted = threading.Event()
        self.prompted_names: list[str] = []
        self.prompted_names_lock = threading.Lock()
        self.is_running = threading.Event()

    def run(self) -> None:
        """Loops forever until stopped. The loop blocks on waiting
        for the 'is_prompted' event to be set, then calls
        :func:`~Follower.pull_and_process` method for each
        prompted name.
        """
        self.has_started.set()

        try:
            while not self.is_stopping.is_set():
                self.is_prompted.wait()

                with self.prompted_names_lock:
                    prompted_names = self.prompted_names
                    self.prompted_names = []
                    self.is_prompted.clear()
                for name in prompted_names:
                    self.follower.pull_and_process(name)
        except Exception as e:
            self.error = EventProcessingError(str(e))
            self.error.__cause__ = e
            self.has_errored.set()

    def receive_recording_event(
        self, new_recording_event: RecordingEvent[TAggregateID]
    ) -> None:
        """Receives prompt by appending name of
        leader to list of prompted names.
        """
        leader_name = new_recording_event.application_name
        with self.prompted_names_lock:
            if leader_name not in self.prompted_names:
                self.prompted_names.append(leader_name)
            self.is_prompted.set()

    def stop(self) -> None:
        self.is_stopping.set()
        self.is_prompted.set()


class NewMultiThreadedRunner(
    Runner[TAggregateID], RecordingEventReceiver[TAggregateID]
):
    """Runs a :class:`System` with multiple threads in a new way."""

    QUEUE_MAX_SIZE: int = 0

    def __init__(
        self,
        system: System,
        env: EnvType | None = None,
    ):
        """Initialises runner with the given :class:`System`."""
        super().__init__(system=system, env=env)
        self.apps: dict[str, Application[TAggregateID]] = {}
        self.pulling_threads: dict[str, list[PullingThread[TAggregateID]]] = {}
        self.processing_queues: dict[
            str, Queue[list[ProcessingJob[TAggregateID]] | None]
        ] = {}
        self.all_threads: list[
            PullingThread[TAggregateID]
            | ConvertingThread[TAggregateID]
            | ProcessingThread[TAggregateID]
        ] = []
        self.has_errored = threading.Event()

        # Construct followers.
        for follower_name in self.system.followers:
            follower_class = self.system.follower_cls(follower_name)
            try:
                follower = follower_class(env=self.env)
            except Exception:
                self.has_errored.set()
                raise
            self.apps[follower_name] = follower

        # Construct non-follower leaders.
        for leader_name in self.system.leaders_only:
            self.apps[leader_name] = self.system.leader_cls(leader_name)(env=self.env)

        # Construct singles.
        for name in self.system.singles:
            single = self.system.get_app_cls(name)(env=self.env)
            self.apps[name] = single

    def start(self) -> None:
        """Starts the runner.

        A multi-threaded runner thread is started for each
        'follower' application in the system, and constructs
        an instance of each non-follower leader application in
        the system. The followers are then setup to follow the
        applications they follow (have a notification log reader
        with the notification log of the leader), and their leaders
        are  setup to lead the follower's thead (send prompts).
        """
        super().start()

        # Start the processing threads.
        for follower_name in self.system.followers:
            follower = cast(Follower[Any], self.apps[follower_name])
            processing_queue: Queue[list[ProcessingJob[TAggregateID]] | None] = Queue(
                maxsize=self.QUEUE_MAX_SIZE
            )
            self.processing_queues[follower_name] = processing_queue
            processing_thread = ProcessingThread(
                processing_queue=processing_queue,
                follower=follower,
                has_errored=self.has_errored,
            )
            self.all_threads.append(processing_thread)
            processing_thread.start()

        for edge in self.system.edges:
            # Set up follower to pull notifications from leader.
            leader_name = edge[0]
            leader = cast("Leader[Any]", self.apps[leader_name])
            follower_name = edge[1]
            follower = cast(Follower[Any], self.apps[follower_name])
            follower.follow(leader.name, leader.notification_log)

            # Create converting queue.
            converting_queue: Queue[ConvertingJob[TAggregateID]] = Queue(
                maxsize=self.QUEUE_MAX_SIZE
            )

            # Start converting thread.
            converting_thread = ConvertingThread(
                converting_queue=converting_queue,
                processing_queue=self.processing_queues[follower_name],
                follower=follower,
                leader_name=leader_name,
                has_errored=self.has_errored,
            )
            self.all_threads.append(converting_thread)
            converting_thread.start()

            # Start pulling thread.
            pulling_thread = PullingThread(
                converting_queue=converting_queue,
                follower=follower,
                leader_name=leader_name,
                has_errored=self.has_errored,
            )
            self.all_threads.append(pulling_thread)
            pulling_thread.start()
            if leader_name not in self.pulling_threads:
                self.pulling_threads[leader_name] = []
            self.pulling_threads[leader_name].append(pulling_thread)

        # Wait until all the threads have started.
        for thread in self.all_threads:
            thread.has_started.wait()

        # Subscribe for notifications from leaders.
        for leader_name in self.system.leaders:
            leader = cast("Leader[Any]", self.apps[leader_name])
            assert isinstance(leader, Leader)
            leader.lead(self)

    def watch_for_errors(self, timeout: float | None = None) -> bool:
        if self.has_errored.wait(timeout=timeout):
            self.stop()
        return self.has_errored.is_set()

    def stop(self) -> None:
        for thread in self.all_threads:
            thread.stop()
        for thread in self.all_threads:
            thread.join(timeout=2)
        for app in self.apps.values():
            app.close()
        self.apps.clear()
        self.reraise_thread_errors()

    def reraise_thread_errors(self) -> None:
        for thread in self.all_threads:
            if thread.error:
                raise thread.error

    def get(self, cls: type[TApplication]) -> TApplication:
        app = self.apps[cls.name]
        assert isinstance(app, cls)
        return app

    def receive_recording_event(
        self, new_recording_event: RecordingEvent[TAggregateID]
    ) -> None:
        for pulling_thread in self.pulling_threads[
            new_recording_event.application_name
        ]:
            pulling_thread.receive_recording_event(new_recording_event)


class PullingThread(threading.Thread, Generic[TAggregateID]):
    """Receives or pulls notifications from the given leader, and
    puts them on a queue for conversion into processing jobs.
    """

    def __init__(
        self,
        converting_queue: Queue[ConvertingJob[TAggregateID]],
        follower: Follower[Any],
        leader_name: str,
        has_errored: threading.Event,
    ):
        super().__init__(daemon=True)
        self.overflow_event = threading.Event()
        self.recording_event_queue: Queue[RecordingEvent[TAggregateID] | None] = Queue(
            maxsize=100
        )
        self.converting_queue: Queue[ConvertingJob[TAggregateID]] = converting_queue
        self.receive_lock = threading.Lock()
        self.follower = follower
        self.leader_name = leader_name
        self.error: Exception | None = None
        self.has_errored = has_errored
        self.is_stopping = threading.Event()
        self.has_started = threading.Event()
        self.mapper = self.follower.mappers[self.leader_name]
        self.previous_max_notification_id = self.follower.recorder.max_tracking_id(
            application_name=self.leader_name
        )

    def run(self) -> None:
        self.has_started.set()
        try:
            while not self.is_stopping.is_set():
                recording_event = self.recording_event_queue.get()
                self.recording_event_queue.task_done()
                if recording_event is None:
                    return
                # Ignore recording event if already seen a subsequent.
                if (
                    recording_event.previous_max_notification_id is not None
                    and self.previous_max_notification_id is not None
                    and recording_event.previous_max_notification_id
                    < self.previous_max_notification_id
                ):
                    continue

                # Catch up if there is a gap in sequence of recording events.
                if (
                    recording_event.previous_max_notification_id is None
                    or self.previous_max_notification_id is None
                    or recording_event.previous_max_notification_id
                    > self.previous_max_notification_id
                ):
                    start = self.previous_max_notification_id
                    stop = recording_event.recordings[0].notification.id - 1
                    for notifications in self.follower.pull_notifications(
                        self.leader_name,
                        start=start,
                        stop=stop,
                        inclusive_of_start=False,
                    ):
                        self.converting_queue.put(notifications)
                        self.previous_max_notification_id = notifications[-1].id
                self.converting_queue.put(recording_event)
                self.previous_max_notification_id = recording_event.recordings[
                    -1
                ].notification.id
        except Exception as e:
            self.error = NotificationPullingError(str(e))
            self.error.__cause__ = e
            self.has_errored.set()

    def receive_recording_event(
        self, recording_event: RecordingEvent[TAggregateID]
    ) -> None:
        try:
            self.recording_event_queue.put(recording_event, timeout=0)
        except Full:
            self.overflow_event.set()

    def stop(self) -> None:
        self.is_stopping.set()
        self.recording_event_queue.put(None)


class ConvertingThread(threading.Thread, Generic[TAggregateID]):
    """Converts notifications into processing jobs."""

    def __init__(
        self,
        converting_queue: Queue[ConvertingJob[TAggregateID]],
        processing_queue: Queue[list[ProcessingJob[TAggregateID]] | None],
        follower: Follower[Any],
        leader_name: str,
        has_errored: threading.Event,
    ):
        super().__init__(daemon=True)
        self.converting_queue: Queue[ConvertingJob[TAggregateID]] = converting_queue
        self.processing_queue: Queue[list[ProcessingJob[TAggregateID]] | None] = (
            processing_queue
        )
        self.follower = follower
        self.leader_name = leader_name
        self.error: Exception | None = None
        self.has_errored = has_errored
        self.is_stopping = threading.Event()
        self.has_started = threading.Event()
        self.mapper = self.follower.mappers[self.leader_name]

    def run(self) -> None:
        self.has_started.set()
        try:
            while True:
                recording_event_or_notifications = self.converting_queue.get()
                self.converting_queue.task_done()
                if (
                    self.is_stopping.is_set()
                    or recording_event_or_notifications is None
                ):
                    return

                processing_jobs = []

                if isinstance(recording_event_or_notifications, RecordingEvent):
                    recording_event = recording_event_or_notifications
                    for recording in recording_event.recordings:
                        if (
                            self.follower.topics
                            and recording.notification.topic not in self.follower.topics
                        ):
                            continue
                        tracking = Tracking(
                            application_name=recording_event.application_name,
                            notification_id=recording.notification.id,
                        )
                        processing_jobs.append((recording.domain_event, tracking))
                else:
                    notifications = recording_event_or_notifications
                    processing_jobs = self.follower.convert_notifications(
                        leader_name=self.leader_name, notifications=notifications
                    )
                if processing_jobs:
                    self.processing_queue.put(processing_jobs)
        except Exception as e:
            print(traceback.format_exc())  # noqa: T201
            self.error = NotificationConvertingError(str(e))
            self.error.__cause__ = e
            self.has_errored.set()

    def stop(self) -> None:
        self.is_stopping.set()
        self.converting_queue.put(None)


class ProcessingThread(threading.Thread, Generic[TAggregateID]):
    """A processing thread gets events from a processing queue, and
    calls the application's process_event() method.
    """

    def __init__(
        self,
        processing_queue: Queue[list[ProcessingJob[TAggregateID]] | None],
        follower: Follower[Any],
        has_errored: threading.Event,
    ):
        super().__init__(daemon=True)
        self.processing_queue: Queue[list[ProcessingJob[TAggregateID]] | None] = (
            processing_queue
        )
        self.follower = follower
        self.error: Exception | None = None
        self.has_errored = has_errored
        self.is_stopping = threading.Event()
        self.has_started = threading.Event()

    def run(self) -> None:
        self.has_started.set()
        try:
            while True:
                jobs = self.processing_queue.get()
                self.processing_queue.task_done()
                if self.is_stopping.is_set() or jobs is None:
                    return
                for domain_event, tracking in jobs:
                    self.follower.process_event(domain_event, tracking)
        except Exception as e:
            self.error = EventProcessingError(str(e))
            self.error.__cause__ = e
            self.has_errored.set()

    def stop(self) -> None:
        self.is_stopping.set()
        self.processing_queue.put(None)


class NotificationLogReader:
    """Reads domain event notifications from a notification log."""

    DEFAULT_SECTION_SIZE = 10

    def __init__(
        self,
        notification_log: NotificationLog,
        section_size: int = DEFAULT_SECTION_SIZE,
    ):
        """Initialises a reader with the given notification log,
        and optionally a section size integer which determines
        the requested number of domain event notifications in
        each section retrieved from the notification log.
        """
        self.notification_log = notification_log
        self.section_size = section_size

    def read(self, *, start: int) -> Iterator[Notification]:
        """Returns a generator that yields event notifications
        from the reader's notification log, starting from
        given start position (a notification ID).

        This method traverses the linked list of sections presented by
        a notification log, and yields the individual event notifications
        that are contained in each section. When all the event notifications
        from a section have been yielded, the reader will retrieve the next
        section, and continues yielding event notification until all subsequent
        event notifications in the notification log from the start position
        have been yielded.
        """
        section_id = f"{start},{start + self.section_size - 1}"
        while True:
            section: Section = self.notification_log[section_id]
            yield from section.items
            if section.next_id is None:
                break
            else:
                section_id = section.next_id

    def select(
        self,
        *,
        start: int | None,
        stop: int | None = None,
        topics: Sequence[str] = (),
        inclusive_of_start: bool = True,
    ) -> Iterator[Sequence[Notification]]:
        """Returns a generator that yields lists of event notifications
        from the reader's notification log, starting from given start
        position (a notification ID).

        This method selects a limited list of notifications from a
        notification log and yields event notifications in batches.
        When one list of event notifications has been yielded,
        the reader will retrieve another list, and continue until
        all subsequent event notifications in the notification log
        from the start position have been yielded.
        """
        while True:
            notifications = self.notification_log.select(
                start=start,
                stop=stop,
                limit=self.section_size,
                topics=topics,
                inclusive_of_start=inclusive_of_start,
            )
            # Stop if zero notifications.
            if len(notifications) == 0:
                break

            # Otherwise, yield and continue.
            start = notifications[-1].id
            if inclusive_of_start:
                start += 1
            yield notifications



from __future__ import annotations

import contextlib
import os
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Iterator, Sequence
from copy import deepcopy
from dataclasses import dataclass
from itertools import chain
from threading import Event, Lock
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    TypeVar,
    cast,
)
from warnings import warn

from eventsourcing.domain import (
    Aggregate,
    BaseAggregate,
    CanMutateProtocol,
    CollectEventsProtocol,
    DomainEventProtocol,
    EventSourcingError,
    MutableOrImmutableAggregate,
    SDomainEvent,
    SnapshotProtocol,
    TAggregateID,
    TDomainEvent,
    TMutableOrImmutableAggregate,
    datetime_now_with_tzinfo,
)
from eventsourcing.persistence import (
    ApplicationRecorder,
    DatetimeAsISO,
    DecimalAsStr,
    EventStore,
    InfrastructureFactory,
    JSONTranscoder,
    Mapper,
    Notification,
    Recording,
    Tracking,
    Transcoder,
    UUIDAsHex,
)
from eventsourcing.utils import Environment, EnvType, strtobool

if TYPE_CHECKING:
    from types import TracebackType
    from uuid import UUID

    from typing_extensions import Self

ProjectorFunction = Callable[
    [TMutableOrImmutableAggregate | None, Iterable[TDomainEvent]],
    TMutableOrImmutableAggregate | None,
]

MutatorFunction = Callable[
    [TDomainEvent, TMutableOrImmutableAggregate | None],
    TMutableOrImmutableAggregate | None,
]


class ProgrammingError(Exception):
    pass


def project_aggregate(
    aggregate: TMutableOrImmutableAggregate | None,
    domain_events: Iterable[DomainEventProtocol[Any]],
) -> TMutableOrImmutableAggregate | None:
    """Projector function for aggregate projections, which works
    by successively calling aggregate mutator function mutate()
    on each of the given list of domain events in turn.
    """
    for domain_event in domain_events:
        assert isinstance(domain_event, CanMutateProtocol)
        aggregate = domain_event.mutate(aggregate)
    return aggregate


S = TypeVar("S")
T = TypeVar("T")


class Cache(Generic[S, T]):
    def __init__(self) -> None:
        self.cache: dict[S, Any] = {}

    def get(self, key: S, *, evict: bool = False) -> T:
        if evict:
            return self.cache.pop(key)
        return self.cache[key]

    def put(self, key: S, value: T | None) -> None:
        if value is not None:
            self.cache[key] = value


class LRUCache(Cache[S, T]):
    """Size limited caching that tracks accesses by recency.

    This is basically copied from functools.lru_cache. But
    we need to know when there was a cache hit, so we can
    fast-forward the aggregate with new stored events.
    """

    sentinel = object()  # unique object used to signal cache misses
    PREV, NEXT, KEY, RESULT = 0, 1, 2, 3  # names for the link fields

    def __init__(self, maxsize: int):
        # Constants shared by all lru cache instances:
        super().__init__()
        self.maxsize = maxsize
        self.full = False
        self.lock = Lock()  # because linkedlist updates aren't threadsafe
        self.root: list[Any] = []  # root of the circular doubly linked list
        self.clear()

    def clear(self) -> None:
        self.root[:] = [
            self.root,
            self.root,
            None,
            None,
        ]  # initialize by pointing to self

    def get(self, key: S, *, evict: bool = False) -> T:
        with self.lock:
            link = self.cache.get(key)
            if link is not None:
                link_prev, link_next, _key, result = link
                if not evict:
                    # Move the link to the front of the circular queue.
                    link_prev[self.NEXT] = link_next
                    link_next[self.PREV] = link_prev
                    last = self.root[self.PREV]
                    last[self.NEXT] = self.root[self.PREV] = link
                    link[self.PREV] = last
                    link[self.NEXT] = self.root
                else:
                    # Remove the link.
                    link_prev[self.NEXT] = link_next
                    link_next[self.PREV] = link_prev
                    del self.cache[key]
                    self.full = self.cache.__len__() >= self.maxsize

                return result
            raise KeyError

    def put(self, key: S, value: T | None) -> Any | None:
        evicted_key = None
        evicted_value = None
        with self.lock:
            link = self.cache.get(key)
            if link is not None:
                # Set value.
                link[self.RESULT] = value
                # Move the link to the front of the circular queue.
                link_prev, link_next, _key, _ = link
                link_prev[self.NEXT] = link_next
                link_next[self.PREV] = link_prev
                last = self.root[self.PREV]
                last[self.NEXT] = self.root[self.PREV] = link
                link[self.PREV] = last
                link[self.NEXT] = self.root
            elif self.full:
                # Use the old root to store the new key and result.
                oldroot = self.root
                oldroot[self.KEY] = key
                oldroot[self.RESULT] = value
                # Empty the oldest link and make it the new root.
                # Keep a reference to the old key and old result to
                # prevent their ref counts from going to zero during the
                # update. That will prevent potentially arbitrary object
                # clean-up code (i.e. __del__) from running while we're
                # still adjusting the links.
                self.root = oldroot[self.NEXT]
                evicted_key = self.root[self.KEY]
                evicted_value = self.root[self.RESULT]
                self.root[self.KEY] = self.root[self.RESULT] = None
                # Now update the cache dictionary.
                del self.cache[evicted_key]
                # Save the potentially reentrant cache[key] assignment
                # for last, after the root and links have been put in
                # a consistent state.
                self.cache[key] = oldroot
            else:
                # Put result in a new link at the front of the queue.
                last = self.root[self.PREV]
                link = [last, self.root, key, value]
                last[self.NEXT] = self.root[self.PREV] = self.cache[key] = link
                # Use the __len__() bound method instead of the len() function
                # which could potentially be wrapped in an lru_cache itself.
                self.full = self.cache.__len__() >= self.maxsize
        return evicted_key, evicted_value


class Repository(Generic[TAggregateID]):
    """Reconstructs aggregates from events in an
    :class:`~eventsourcing.persistence.EventStore`,
    possibly using snapshot store to avoid replaying
    all events.
    """

    FASTFORWARD_LOCKS_CACHE_MAXSIZE = 50

    def __init__(
        self,
        event_store: EventStore[TAggregateID],
        *,
        snapshot_store: EventStore[TAggregateID] | None = None,
        cache_maxsize: int | None = None,
        fastforward: bool = True,
        fastforward_skipping: bool = False,
        deepcopy_from_cache: bool = True,
    ):
        """Initialises repository with given event store (an
        :class:`~eventsourcing.persistence.EventStore` for aggregate
        :class:`~eventsourcing.domain.AggregateEvent` objects)
        and optionally a snapshot store (an
        :class:`~eventsourcing.persistence.EventStore` for aggregate
        :class:`~eventsourcing.domain.Snapshot` objects).
        """
        self.event_store: EventStore[TAggregateID] = event_store
        self.snapshot_store: EventStore[TAggregateID] | None = snapshot_store

        if cache_maxsize is None:
            self.cache: (
                Cache[TAggregateID, MutableOrImmutableAggregate[TAggregateID]] | None
            ) = None
        elif cache_maxsize <= 0:
            self.cache = Cache()
        else:
            self.cache = LRUCache(maxsize=cache_maxsize)
        self.fastforward = fastforward
        self.fastforward_skipping = fastforward_skipping
        self.deepcopy_from_cache = deepcopy_from_cache

        # Because fast-forwarding a cached aggregate isn't thread-safe.
        self._fastforward_locks_lock = Lock()
        self._fastforward_locks_cache: LRUCache[TAggregateID, Lock] = LRUCache(
            maxsize=self.FASTFORWARD_LOCKS_CACHE_MAXSIZE
        )
        self._fastforward_locks_inuse: dict[TAggregateID, tuple[Lock, int]] = {}

    def get(
        self,
        aggregate_id: TAggregateID,
        *,
        version: int | None = None,
        projector_func: ProjectorFunction[
            TMutableOrImmutableAggregate, TDomainEvent
        ] = project_aggregate,
        fastforward_skipping: bool = False,
        deepcopy_from_cache: bool = True,
    ) -> TMutableOrImmutableAggregate:
        """Reconstructs an :class:`~eventsourcing.domain.Aggregate` for a
        given ID from stored events, optionally at a particular version.
        """
        if self.cache and version is None:
            try:
                # Look for aggregate in the cache.
                aggregate = cast(
                    "TMutableOrImmutableAggregate", self.cache.get(aggregate_id)
                )
            except KeyError:
                # Reconstruct aggregate from stored events.
                aggregate = self._reconstruct_aggregate(
                    aggregate_id, None, projector_func
                )
                # Put aggregate in the cache.
                self.cache.put(aggregate_id, aggregate)
            else:
                if self.fastforward:
                    # Fast-forward cached aggregate.
                    fastforward_lock = self._use_fastforward_lock(aggregate_id)
                    # TODO: Should this be 'fastforward or self.fastforward_skipping'?
                    blocking = not (fastforward_skipping or self.fastforward_skipping)
                    try:
                        if fastforward_lock.acquire(blocking=blocking):
                            try:
                                new_events = self.event_store.get(
                                    originator_id=aggregate_id, gt=aggregate.version
                                )
                                _aggregate = projector_func(
                                    aggregate,
                                    cast(
                                        "Iterable[TDomainEvent]",
                                        new_events,
                                    ),
                                )
                                if _aggregate is None:
                                    raise AggregateNotFoundError(aggregate_id)
                                aggregate = _aggregate
                            finally:
                                fastforward_lock.release()
                    finally:
                        self._disuse_fastforward_lock(aggregate_id)

            # Copy mutable aggregates for commands, so bad mutations don't corrupt.
            if deepcopy_from_cache and self.deepcopy_from_cache:
                aggregate = deepcopy(aggregate)
        else:
            # Reconstruct historical version of aggregate from stored events.
            aggregate = self._reconstruct_aggregate(
                aggregate_id, version, projector_func
            )
        return aggregate

    def _reconstruct_aggregate(
        self,
        aggregate_id: TAggregateID,
        version: int | None,
        projector_func: ProjectorFunction[TMutableOrImmutableAggregate, TDomainEvent],
    ) -> TMutableOrImmutableAggregate:
        gt: int | None = None

        if self.snapshot_store is not None:
            # Try to get a snapshot.
            snapshots = list(
                self.snapshot_store.get(
                    originator_id=aggregate_id,
                    desc=True,
                    limit=1,
                    lte=version,
                ),
            )
            if snapshots:
                gt = snapshots[0].originator_version
        else:
            snapshots = []

        # Get aggregate events.
        aggregate_events = self.event_store.get(
            originator_id=aggregate_id,
            gt=gt,
            lte=version,
        )

        # Reconstruct the aggregate from its events.
        initial: TMutableOrImmutableAggregate | None = None
        aggregate = projector_func(
            initial,
            chain(
                cast("Iterable[TDomainEvent]", snapshots),
                cast("Iterable[TDomainEvent]", aggregate_events),
            ),
        )

        # Raise exception if "not found".
        if aggregate is None:
            msg = f"Aggregate {aggregate_id!r} version {version!r} not found."
            raise AggregateNotFoundError(msg)
        # Return the aggregate.
        return aggregate

    def _use_fastforward_lock(self, aggregate_id: TAggregateID) -> Lock:
        lock: Lock | None = None
        with self._fastforward_locks_lock:
            num_users = 0
            with contextlib.suppress(KeyError):
                lock, num_users = self._fastforward_locks_inuse[aggregate_id]
            if lock is None:
                with contextlib.suppress(KeyError):
                    lock = self._fastforward_locks_cache.get(aggregate_id, evict=True)
            if lock is None:
                lock = Lock()
            num_users += 1
            self._fastforward_locks_inuse[aggregate_id] = (lock, num_users)
            return lock

    def _disuse_fastforward_lock(self, aggregate_id: TAggregateID) -> None:
        with self._fastforward_locks_lock:
            lock_, num_users = self._fastforward_locks_inuse[aggregate_id]
            num_users -= 1
            if num_users == 0:
                del self._fastforward_locks_inuse[aggregate_id]
                self._fastforward_locks_cache.put(aggregate_id, lock_)
            else:
                self._fastforward_locks_inuse[aggregate_id] = (lock_, num_users)

    def __contains__(self, item: TAggregateID) -> bool:
        """Tests to see if an aggregate exists in the repository."""
        try:
            self.get(aggregate_id=item)
        except AggregateNotFoundError:
            return False
        else:
            return True


@dataclass(frozen=True)
class Section:
    """Frozen dataclass that represents a section from a :class:`NotificationLog`.
    The :data:`items` attribute contains a list of
    :class:`~eventsourcing.persistence.Notification` objects.
    The :data:`id` attribute is the section ID, two integers
    separated by a comma that described the first and last
    notification ID that are included in the section.
    The :data:`next_id` attribute describes the section ID
    of the next section, and will be set if the section contains
    as many notifications as were requested.

    Constructor arguments:

    :param Optional[str] id: section ID of this section e.g. "1,10"
    :param list[Notification] items: a list of event notifications
    :param Optional[str] next_id: section ID of the following section
    """

    id: str | None
    items: Sequence[Notification]
    next_id: str | None


class NotificationLog(ABC):
    """Abstract base class for notification logs."""

    @abstractmethod
    def __getitem__(self, section_id: str) -> Section:
        """Returns a :class:`Section` of
        :class:`~eventsourcing.persistence.Notification` objects
        from the notification log.
        """

    @abstractmethod
    def select(
        self,
        start: int | None,
        limit: int,
        stop: int | None = None,
        topics: Sequence[str] = (),
        *,
        inclusive_of_start: bool = True,
    ) -> Sequence[Notification]:
        """Returns a selection of
        :class:`~eventsourcing.persistence.Notification` objects
        from the notification log.
        """


class LocalNotificationLog(NotificationLog):
    """Notification log that presents sections of event notifications
    retrieved from an :class:`~eventsourcing.persistence.ApplicationRecorder`.
    """

    DEFAULT_SECTION_SIZE = 10

    def __init__(
        self,
        recorder: ApplicationRecorder,
        section_size: int = DEFAULT_SECTION_SIZE,
    ):
        """Initialises a local notification object with given
        :class:`~eventsourcing.persistence.ApplicationRecorder`
        and an optional section size.

        Constructor arguments:

        :param ApplicationRecorder recorder: application recorder from which event
            notifications will be selected
        :param int section_size: number of notifications to include in a section

        """
        self.recorder = recorder
        self.section_size = section_size

    def __getitem__(self, requested_section_id: str) -> Section:
        """Returns a :class:`Section` of event notifications
        based on the requested section ID. The section ID of
        the returned section will describe the event
        notifications that are actually contained in
        the returned section, and may vary from the
        requested section ID if there are fewer notifications
        in the recorder than were requested, or if there
        are gaps in the sequence of recorded event notification.
        """
        # Interpret the section ID.
        parts = requested_section_id.split(",")
        part1 = int(parts[0])
        part2 = int(parts[1])
        start = max(1, part1)
        limit = min(max(0, part2 - start + 1), self.section_size)

        # Select notifications.
        notifications = self.select(start, limit)

        # Get next section ID.
        actual_section_id: str | None
        next_id: str | None
        if len(notifications):
            last_notification_id = notifications[-1].id
            actual_section_id = self.format_section_id(
                notifications[0].id, last_notification_id
            )
            if len(notifications) == limit:
                next_id = self.format_section_id(
                    last_notification_id + 1, last_notification_id + limit
                )
            else:
                next_id = None
        else:
            actual_section_id = None
            next_id = None

        # Return a section of the notification log.
        return Section(
            id=actual_section_id,
            items=notifications,
            next_id=next_id,
        )

    def select(
        self,
        start: int | None,
        limit: int,
        stop: int | None = None,
        topics: Sequence[str] = (),
        *,
        inclusive_of_start: bool = True,
    ) -> Sequence[Notification]:
        """Returns a selection of
        :class:`~eventsourcing.persistence.Notification` objects
        from the notification log.
        """
        if limit > self.section_size:
            msg = (
                f"Requested limit {limit} greater than section size {self.section_size}"
            )
            raise ValueError(msg)
        return self.recorder.select_notifications(
            start=start,
            limit=limit,
            stop=stop,
            topics=topics,
            inclusive_of_start=inclusive_of_start,
        )

    @staticmethod
    def format_section_id(first_id: int, last_id: int) -> str:
        return f"{first_id},{last_id}"


class ProcessingEvent(Generic[TAggregateID]):
    """Keeps together a :class:`~eventsourcing.persistence.Tracking`
    object, which represents the position of a domain event notification
    in the notification log of a particular application, and the
    new domain events that result from processing that notification.
    """

    def __init__(self, tracking: Tracking | None = None):
        """Initialises the process event with the given tracking object."""
        self.tracking = tracking
        self.events: list[DomainEventProtocol[TAggregateID]] = []
        self.aggregates: dict[
            TAggregateID, MutableOrImmutableAggregate[TAggregateID]
        ] = {}
        self.saved_kwargs: dict[Any, Any] = {}

    def collect_events(
        self,
        *objs: MutableOrImmutableAggregate[TAggregateID]
        | DomainEventProtocol[TAggregateID]
        | None,
        **kwargs: Any,
    ) -> None:
        """Collects pending domain events from the given aggregate."""
        for obj in objs:
            if obj is None:
                continue
            if isinstance(obj, DomainEventProtocol):
                self.events.append(obj)
            else:
                if isinstance(obj, CollectEventsProtocol):
                    for event in obj.collect_events():
                        self.events.append(event)
                self.aggregates[obj.id] = obj

        self.saved_kwargs.update(kwargs)

    def save(
        self,
        *aggregates: MutableOrImmutableAggregate[TAggregateID]
        | DomainEventProtocol[TAggregateID]
        | None,
        **kwargs: Any,
    ) -> None:
        warn(
            "'save()' is deprecated, use 'collect_events()' instead",
            DeprecationWarning,
            stacklevel=2,
        )

        self.collect_events(*aggregates, **kwargs)


class Application(Generic[TAggregateID]):
    """Base class for event-sourced applications."""

    name = "Application"
    env: ClassVar[dict[str, str]] = {}
    is_snapshotting_enabled: bool = False
    snapshotting_intervals: ClassVar[
        dict[type[MutableOrImmutableAggregate[Any]], int]
    ] = {}
    snapshotting_projectors: ClassVar[
        dict[
            type[MutableOrImmutableAggregate[Any]],
            ProjectorFunction[Any, Any],
        ]
    ] = {}
    snapshot_class: type[SnapshotProtocol[TAggregateID]] | None = None
    log_section_size = 10
    notify_topics: Sequence[str] = []

    AGGREGATE_CACHE_MAXSIZE = "AGGREGATE_CACHE_MAXSIZE"
    AGGREGATE_CACHE_FASTFORWARD = "AGGREGATE_CACHE_FASTFORWARD"
    AGGREGATE_CACHE_FASTFORWARD_SKIPPING = "AGGREGATE_CACHE_FASTFORWARD_SKIPPING"
    DEEPCOPY_FROM_AGGREGATE_CACHE = "DEEPCOPY_FROM_AGGREGATE_CACHE"

    def __init_subclass__(cls, **kwargs: Any) -> None:
        if "name" not in cls.__dict__:
            cls.name = cls.__name__

    def __init__(self, env: EnvType | None = None) -> None:
        """Initialises an application with an
        :class:`~eventsourcing.persistence.InfrastructureFactory`,
        a :class:`~eventsourcing.persistence.Mapper`,
        an :class:`~eventsourcing.persistence.ApplicationRecorder`,
        an :class:`~eventsourcing.persistence.EventStore`,
        a :class:`~eventsourcing.application.Repository`, and
        a :class:`~eventsourcing.application.LocalNotificationLog`.
        """
        self.closing = Event()
        self.env = self.construct_env(self.name, env)  # type: ignore[misc]
        self.factory = self.construct_factory(self.env)
        self.mapper: Mapper[TAggregateID] = self.construct_mapper()
        self.recorder = self.construct_recorder()
        self.events: EventStore[TAggregateID] = self.construct_event_store()
        self.snapshots: EventStore[TAggregateID] | None = None
        if self.factory.is_snapshotting_enabled():
            self.snapshots = self.construct_snapshot_store()
        self._repository: Repository[TAggregateID] = self.construct_repository()
        self._notification_log = self.construct_notification_log()

    @property
    def repository(self) -> Repository[TAggregateID]:
        """An application's repository reconstructs aggregates from stored events."""
        return self._repository

    @property
    def notification_log(self) -> LocalNotificationLog:
        """An application's notification log presents all the aggregate events
        of an application in the order they were recorded as a sequence of event
        notifications.
        """
        return self._notification_log

    @property
    def log(self) -> LocalNotificationLog:
        warn(
            "'log' is deprecated, use 'notification_log' instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._notification_log

    def construct_env(self, name: str, env: EnvType | None = None) -> Environment:
        """Constructs environment from which application will be configured."""
        _env = dict(type(self).env)
        if type(self).is_snapshotting_enabled or type(self).snapshotting_intervals:
            _env["IS_SNAPSHOTTING_ENABLED"] = "y"
        _env.update(os.environ)
        if env is not None:
            _env.update(env)
        return Environment(name, _env)

    def construct_factory(self, env: Environment) -> InfrastructureFactory:
        """Constructs an :class:`~eventsourcing.persistence.InfrastructureFactory`
        for use by the application.
        """
        return InfrastructureFactory.construct(env)

    def construct_mapper(self) -> Mapper[TAggregateID]:
        """Constructs a :class:`~eventsourcing.persistence.Mapper`
        for use by the application.
        """
        return self.factory.mapper(transcoder=self.construct_transcoder())

    def construct_transcoder(self) -> Transcoder:
        """Constructs a :class:`~eventsourcing.persistence.Transcoder`
        for use by the application.
        """
        transcoder = self.factory.transcoder()
        if isinstance(transcoder, JSONTranscoder):
            self.register_transcodings(transcoder)
        return transcoder

    def register_transcodings(self, transcoder: JSONTranscoder) -> None:
        """Registers :class:`~eventsourcing.persistence.Transcoding`
        objects on given :class:`~eventsourcing.persistence.JSONTranscoder`.
        """
        transcoder.register(UUIDAsHex())
        transcoder.register(DecimalAsStr())
        transcoder.register(DatetimeAsISO())

    def construct_recorder(self) -> ApplicationRecorder:
        """Constructs an :class:`~eventsourcing.persistence.ApplicationRecorder`
        for use by the application.
        """
        return self.factory.application_recorder()

    def construct_event_store(self) -> EventStore[TAggregateID]:
        """Constructs an :class:`~eventsourcing.persistence.EventStore`
        for use by the application to store and retrieve aggregate
        :class:`~eventsourcing.domain.AggregateEvent` objects.
        """
        return self.factory.event_store(
            mapper=self.mapper,
            recorder=self.recorder,
        )

    def construct_snapshot_store(self) -> EventStore[TAggregateID]:
        """Constructs an :py:class:`~eventsourcing.persistence.EventStore`
        for use by the application to store and retrieve aggregate
        :class:`~eventsourcing.domain.Snapshot` objects.
        """
        recorder = self.factory.aggregate_recorder(purpose="snapshots")
        return self.factory.event_store(
            mapper=self.mapper,
            recorder=recorder,
        )

    def construct_repository(self) -> Repository[TAggregateID]:
        """Constructs a :py:class:`Repository` for use by the application."""
        cache_maxsize_envvar = self.env.get(self.AGGREGATE_CACHE_MAXSIZE)
        cache_maxsize = int(cache_maxsize_envvar) if cache_maxsize_envvar else None
        return Repository(
            event_store=self.events,
            snapshot_store=self.snapshots,
            cache_maxsize=cache_maxsize,
            fastforward=strtobool(self.env.get(self.AGGREGATE_CACHE_FASTFORWARD, "y")),
            fastforward_skipping=strtobool(
                self.env.get(self.AGGREGATE_CACHE_FASTFORWARD_SKIPPING, "n")
            ),
            deepcopy_from_cache=strtobool(
                self.env.get(self.DEEPCOPY_FROM_AGGREGATE_CACHE, "y")
            ),
        )

    def construct_notification_log(self) -> LocalNotificationLog:
        """Constructs a :class:`LocalNotificationLog` for use by the application."""
        return LocalNotificationLog(self.recorder, section_size=self.log_section_size)

    def save(
        self,
        *objs: MutableOrImmutableAggregate[TAggregateID]
        | DomainEventProtocol[TAggregateID]
        | None,
        **kwargs: Any,
    ) -> list[Recording[TAggregateID]]:
        """Collects pending events from given aggregates and
        puts them in the application's event store.
        """
        processing_event: ProcessingEvent[TAggregateID] = ProcessingEvent()
        processing_event.collect_events(*objs, **kwargs)
        recordings = self._record(processing_event)
        self._take_snapshots(processing_event)
        self._notify(recordings)
        self.notify(processing_event.events)  # Deprecated.
        return recordings

    def _record(
        self, processing_event: ProcessingEvent[TAggregateID]
    ) -> list[Recording[TAggregateID]]:
        """Records given process event in the application's recorder."""
        recordings = self.events.put(
            processing_event.events,
            tracking=processing_event.tracking,
            **processing_event.saved_kwargs,
        )
        if self.repository.cache and not self.repository.fastforward:
            for aggregate_id, aggregate in processing_event.aggregates.items():
                self.repository.cache.put(aggregate_id, aggregate)
        return recordings

    def _take_snapshots(self, processing_event: ProcessingEvent[TAggregateID]) -> None:
        # Take snapshots using IDs and types.
        if self.snapshots and self.snapshotting_intervals:
            for event in processing_event.events:
                try:
                    aggregate = processing_event.aggregates[event.originator_id]
                except KeyError:
                    continue
                interval = self.snapshotting_intervals.get(type(aggregate))
                if interval is not None and event.originator_version % interval == 0:
                    try:
                        projector_func = self.snapshotting_projectors[type(aggregate)]
                    except KeyError:
                        if not isinstance(event, CanMutateProtocol):
                            msg = (
                                f"Cannot take snapshot for {type(aggregate)} with "
                                "default project_aggregate() function, because its "
                                f"domain event {type(event)} does not implement "
                                "the 'can mutate' protocol (see CanMutateProtocol)."
                                f" Please define application class {type(self)}"
                                " with class variable 'snapshotting_projectors', "
                                f"to be a dict that has {type(aggregate)} as a key "
                                "with the aggregate projector function for "
                                f"{type(aggregate)} as the value for that key."
                            )
                            raise ProgrammingError(msg) from None

                        projector_func = project_aggregate
                    self.take_snapshot(
                        aggregate_id=event.originator_id,
                        version=event.originator_version,
                        projector_func=projector_func,
                    )

    def take_snapshot(
        self,
        aggregate_id: TAggregateID,
        version: int | None = None,
        projector_func: ProjectorFunction[Any, Any] = project_aggregate,
    ) -> None:
        """Takes a snapshot of the recorded state of the aggregate,
        and puts the snapshot in the snapshot store.
        """
        if self.snapshots is None:
            msg = (
                "Can't take snapshot without snapshots store. Please "
                "set environment variable IS_SNAPSHOTTING_ENABLED to "
                "a true value (e.g. 'y'), or set 'is_snapshotting_enabled' "
                "on application class, or set 'snapshotting_intervals' on "
                "application class."
            )
            raise AssertionError(msg)
        aggregate: BaseAggregate[TAggregateID] = self.repository.get(
            aggregate_id, version=version, projector_func=projector_func
        )
        snapshot_class = getattr(type(aggregate), "Snapshot", type(self).snapshot_class)
        if snapshot_class is None:
            msg = (
                "Neither application nor aggregate have a snapshot class. "
                f"Please either define a nested 'Snapshot' class on {type(aggregate)} "
                f"or set class attribute 'snapshot_class' on {type(self)}."
            )
            raise AssertionError(msg)

        snapshot = snapshot_class.take(aggregate)
        self.snapshots.put([snapshot])

    def notify(self, new_events: list[DomainEventProtocol[TAggregateID]]) -> None:
        """Deprecated.

        Called after new aggregate events have been saved. This
        method on this class doesn't actually do anything,
        but this method may be implemented by subclasses that
        need to take action when new domain events have been saved.
        """

    def _notify(self, recordings: list[Recording[TAggregateID]]) -> None:
        """Called after new aggregate events have been saved. This
        method on this class doesn't actually do anything,
        but this method may be implemented by subclasses that
        need to take action when new domain events have been saved.
        """

    def close(self) -> None:
        self.closing.set()
        self.factory.close()

    def __enter__(self) -> Self:
        self.factory.__enter__()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()
        self.factory.__exit__(exc_type, exc_val, exc_tb)

    def __del__(self) -> None:
        with contextlib.suppress(AttributeError):
            self.close()


TApplication = TypeVar("TApplication", bound=Application[Any])


class AggregateNotFoundError(EventSourcingError):
    """Raised when an :class:`~eventsourcing.domain.Aggregate`
    object is not found in a :class:`Repository`.
    """


class EventSourcedLog(Generic[TDomainEvent]):
    """Constructs a sequence of domain events, like an aggregate.
    But unlike an aggregate the events can be triggered
    and selected for use in an application without
    reconstructing a current state from all the events.

    This allows an indefinitely long sequence of events to be
    generated and used without the practical restrictions of
    projecting the events into a current state before they
    can be used, which is useful e.g. for logging and
    progressively discovering all the aggregate IDs of a
    particular type in an application.
    """

    def __init__(
        self,
        events: EventStore[Any],
        originator_id: UUID,
        logged_cls: type[TDomainEvent],  # TODO: Rename to 'event_class' in v10.
    ):
        self.events = events
        self.originator_id = originator_id
        self.logged_cls = logged_cls  # TODO: Rename to 'event_class' in v10.

    def trigger_event(
        self,
        next_originator_version: int | None = None,
        **kwargs: Any,
    ) -> TDomainEvent:
        """Constructs and returns a new log event."""
        return self._trigger_event(
            logged_cls=self.logged_cls,
            next_originator_version=next_originator_version,
            **kwargs,
        )

    def _trigger_event(
        self,
        logged_cls: type[SDomainEvent],
        next_originator_version: int | None = None,
        **kwargs: Any,
    ) -> SDomainEvent:
        """Constructs and returns a new log event."""
        if next_originator_version is None:
            last_logged = self.get_last()
            if last_logged is None:
                next_originator_version = Aggregate.INITIAL_VERSION
            else:
                next_originator_version = last_logged.originator_version + 1

        return logged_cls(
            originator_id=self.originator_id,
            originator_version=next_originator_version,
            timestamp=datetime_now_with_tzinfo(),
            **kwargs,
        )

    def get_first(self) -> TDomainEvent | None:
        """Selects the first logged event."""
        try:
            return next(self.get(limit=1))
        except StopIteration:
            return None

    def get_last(self) -> TDomainEvent | None:
        """Selects the last logged event."""
        try:
            return next(self.get(desc=True, limit=1))
        except StopIteration:
            return None

    def get(
        self,
        *,
        gt: int | None = None,
        lte: int | None = None,
        desc: bool = False,
        limit: int | None = None,
    ) -> Iterator[TDomainEvent]:
        """Selects a range of logged events with limit,
        with ascending or descending order.
        """
        return cast(
            "Iterator[TDomainEvent]",
            self.events.get(
                originator_id=self.originator_id,
                gt=gt,
                lte=lte,
                desc=desc,
                limit=limit,
            ),
        )
